# Copyright (c) 2024, Jiun-Cheng Jiang. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Synchrous processing of quantum circuits with PyTorch.

Features:
    - Single-qubit quantum circuits (Faster than other libraries)
    - Two-qubit quantum circuits
"""

import math

import torch

# Constants
INV_SQRT2 = math.sqrt(2.0) / 2.0  # 1/sqrt(2) for Hadamard gate


class TorchGates:
    @staticmethod
    def identity_gate(shape) -> torch.Tensor:
        """
        shape: (out_dim, in_dim)

        return: torch.Tensor, shape: (2, 2, out_dim, in_dim)
        """
        return torch.stack(
            [
                torch.stack([torch.ones(*shape), torch.zeros(*shape)]),
                torch.stack([torch.zeros(*shape), torch.ones(*shape)]),
            ],
        )

    @staticmethod
    def i_gate(shape):
        """
        alias for identity_gate

        shape: (out_dim, in_dim)

        return: torch.Tensor, shape: (2, 2, out_dim, in_dim)
        """
        pass

    i_gate = identity_gate  # noqa: F811

    @staticmethod
    def rx_gate(theta: torch.Tensor, dtype=torch.complex64) -> torch.Tensor:
        """
        theta: torch.Tensor, shape: (out_dim, in_dim)

        return: torch.Tensor, shape: (2, 2, out_dim, in_dim)
        """
        cos = torch.cos(theta / 2).to(dtype)
        jsin = 1j * torch.sin(-theta / 2)
        return torch.stack(
            [
                torch.stack([cos, jsin]),
                torch.stack([jsin, cos]),
            ],
        )

    @staticmethod
    def ry_gate(theta: torch.Tensor, dtype=torch.complex64) -> torch.Tensor:
        """
        theta: torch.Tensor, shape: (out_dim, in_dim)

        return: torch.Tensor, shape: (2, 2, out_dim, in_dim)
        """
        cos = torch.cos(theta / 2)
        sin = torch.sin(theta / 2)
        return torch.stack(
            [
                torch.stack([cos, -sin]),
                torch.stack([sin, cos]),
            ],
        ).to(dtype)

    @staticmethod
    def rz_gate(theta: torch.Tensor, dtype=torch.complex64) -> torch.Tensor:
        """
        theta: torch.Tensor, shape: (out_dim, in_dim)

        return: torch.Tensor, shape: (2, 2, out_dim, in_dim)
        """
        exp = torch.exp(-0.5j * theta)
        zero = torch.zeros_like(theta)
        return torch.stack(
            [
                torch.stack([exp, zero]),
                torch.stack([zero, torch.conj(exp)]),
            ],
        ).to(dtype)

    @staticmethod
    def h_gate(shape, device, dtype=torch.complex64) -> torch.Tensor:
        """
        shape: (out_dim, in_dim)

        return: torch.Tensor, shape: (2, 2, out_dim, in_dim)
        """
        # Optimize: use pre-computed constant instead of computing 1/sqrt(2) at runtime
        inv_sqrt2 = torch.full(shape, INV_SQRT2, device=device, dtype=dtype)
        return torch.stack(
            [
                torch.stack([inv_sqrt2, inv_sqrt2]),
                torch.stack([inv_sqrt2, -inv_sqrt2]),
            ],
        )

    @staticmethod
    def s_gate(shape) -> torch.Tensor:
        """
        shape: (out_dim, in_dim)

        return: torch.Tensor, shape: (2, 2, out_dim, in_dim)
        """
        return torch.stack(
            [
                torch.stack([torch.ones(*shape), torch.zeros(*shape)]),
                torch.stack([torch.zeros(*shape), 1j * torch.ones(*shape)]),
            ],
        )

    @staticmethod
    def acrx_gate(theta: torch.Tensor, dtype=torch.complex64) -> torch.Tensor:
        """
        Complex extension of RX(acos(theta)) gate.
        *Note: Physically unrealizable.*

        theta: torch.Tensor, shape: (out_dim, in_dim)

        return: torch.Tensor, shape: (2, 2, out_dim, in_dim)
        """

        sq = torch.square(theta).flatten()
        diag = torch.mul(
            torch.sqrt(torch.abs(1 - sq)), torch.where(sq < 1, 1j, 1)
        ).reshape(theta.shape)
        return torch.stack(
            [
                torch.stack([theta, diag]),
                torch.stack([diag, theta]),
            ],
        ).to(dtype)

    @staticmethod
    def tensor_product(gate, another_gate, dtype=None):
        """
        Compute tensor product of two gates.

        Arguments
        ---------
            :gate: torch.Tensor, shape: (2, 2, out_dim, in_dim)
            :another_gate: torch.Tensor, shape: (2, 2, out_dim, in_dim)
            :dtype: torch dtype, optional. If None, uses the dtype of the input gate.
                    Both gates should have the same dtype.

        return: torch.Tensor, shape: (4, 4, out_dim, in_dim)
        """
        if dtype is None:
            dtype = gate.dtype
            # Validate that both gates have the same dtype
            if gate.dtype != another_gate.dtype:
                raise ValueError(
                    f"Gate dtypes must match: got {gate.dtype} and {another_gate.dtype}"
                )
        shape = gate.shape[2:]
        gate = gate.view(2, 2, -1)
        another_gate = another_gate.view(2, 2, -1)
        out = torch.empty(
            4,
            4,
            gate.shape[2],
            dtype=dtype,
            device=gate.device,
        )
        for i in range(out.shape[2]):
            out[:, :, i] = torch.kron(gate[:, :, i], another_gate[:, :, i])
        return out.view(4, 4, *shape)

    @staticmethod
    def cx_gate(shape, control: int, device, dtype=torch.complex64) -> torch.Tensor:
        """
        2-qubits CX (CNOT) gate.

        shape: (out_dim, in_dim)
        control: int

        return: torch.Tensor, shape: (4, 4, out_dim, in_dim)
        """
        assert control in (0, 1), "Control qubit must be 0 or 1."

        gate = torch.zeros(4, 4, *shape, dtype=dtype, device=device)
        gate[0, 0] = 1.0
        gate[1, 1] = 1.0
        gate[2, 3] = 1.0
        gate[3, 2] = 1.0
        if control == 1:
            gate = gate.transpose(0, 1)
        return gate

    @staticmethod
    def cz_gate(shape, device, dtype=torch.complex64) -> torch.Tensor:
        """
        2-qubits CZ gate.

        shape: (out_dim, in_dim)
        control: int

        return: torch.Tensor, shape: (4, 4, out_dim, in_dim)
        """

        gate = torch.zeros(4, 4, *shape, dtype=dtype, device=device)
        gate[0, 0] = 1.0
        gate[1, 1] = 1.0
        gate[2, 2] = 1.0
        gate[3, 3] = -1.0
        return gate


class StateVector:
    """
    1-qubit state vector.

    StateVector.state: torch.Tensor, shape: (batch_size, out_dim, in_dim, 2)
    """

    state: torch.Tensor

    def __init__(
        self,
        batch_size: int,
        out_dim: int,
        in_dim: int,
        device="cpu",
        dtype=torch.complex64,
    ):
        self.device = device
        self.batch_size = batch_size
        self.out_dim = out_dim
        self.in_dim = in_dim
        self.state = torch.zeros(
            batch_size, out_dim, in_dim, 2, dtype=dtype, device=self.device
        )
        self.state[:, :, :, 0] = 1.0
        self.dtype = dtype

    def measure_z(self, fast_measure: bool = True) -> torch.Tensor:
        """
        Measure the state vector in the Z basis.

        Arguments
        ---------
            :fast_measure: bool, default: True. If True, for state |ψ⟩ = α|0⟩ + β|1⟩, return |α| - |β|;
                           if False, return |α|^2 - |β|^2.
                           Which is quantum-inspired method and faster when it is True.
        return: torch.Tensor, shape: (batch_size, out_dim, in_dim)
        """
        return (
            self.state[:, :, :, 0].abs() - self.state[:, :, :, 1].abs()
            if fast_measure
            else torch.square(self.state[:, :, :, 0].abs())
            - torch.square(self.state[:, :, :, 1].abs())
        )

    def measure_x(self, fast_measure: bool = True) -> torch.Tensor:
        """
        Measure the state vector in the X basis.

        Arguments
        ---------
            :fast_measure: bool, default: True. If True, for state |ψ⟩ = α|0⟩ + β|1⟩, return |α| - |β|;
                           if False, return |α|^2 - |β|^2.
                           Which is quantum-inspired method and faster when it is True.
        return: torch.Tensor, shape: (batch_size, out_dim, in_dim)
        """
        tmp_state = StateVector(self.batch_size, self.out_dim, self.in_dim, self.device)
        tmp_state.state.copy_(self.state)
        tmp_state.h()
        return tmp_state.measure_z(fast_measure)

    def measure_y(self, fast_measure: bool = True) -> torch.Tensor:
        """
        Measure the state vector in the Y basis.

        Arguments
        ---------
            :fast_measure: bool, default: True. If True, for state |ψ⟩ = α|0⟩ + β|1⟩, return |α| - |β|;
                           if False, return |α|^2 - |β|^2.
                           Which is quantum-inspired method and faster when it is True.
        return: torch.Tensor, shape: (batch_size, out_dim, in_dim)
        """
        tmp_state = StateVector(self.batch_size, self.out_dim, self.in_dim, self.device)
        tmp_state.state.copy_(self.state)
        tmp_state.s(is_dagger=True)
        tmp_state.h()
        return tmp_state.measure_z(fast_measure)

    def s(self, is_dagger: bool = False):
        """
        Apply Phase gate (or S gate) to the state vector.

        Arguments
        ---------
            :is_dagger: bool, default: False
        """
        gate = TorchGates.s_gate(self.state.shape[1:3]).to(self.device)
        if is_dagger:
            gate = torch.conj_physical(gate).transpose(0, 1)
        self.state = torch.einsum("mnoi,boin->boim", gate, self.state)

    def h(self, is_dagger: bool = False):
        """
        Apply Hadamard gate to the state vector.

        Arguments
        ---------
            :is_dagger: bool, default: False
        """
        gate = TorchGates.h_gate(self.state.shape[1:3], self.device, dtype=self.dtype)
        if is_dagger:
            gate = torch.conj_physical(gate).transpose(0, 1)
        self.state = torch.einsum("mnoi,boin->boim", gate, self.state)

    def x(self):
        """
        Apply Pauli-X gate to the state vector.
        """
        self.state = torch.index_select(
            self.state, dim=-1, index=torch.tensor([1, 0], device=self.device)
        )

    def z(self):
        """
        Apply Pauli-Z gate to the state vector.
        """
        self.state[:, :, :, 1] = -self.state[:, :, :, 1]

    def rx(self, theta: torch.Tensor, is_dagger: bool = False):
        """
        Apply Rotation-X gate to the state vector.

        Arguments
        ---------
            :theta: torch.Tensor, shape: (out_dim, in_dim)
            :is_dagger: bool, default: False
        """
        gate = TorchGates.rx_gate(theta, dtype=self.dtype)
        if is_dagger:
            gate = torch.conj_physical(gate).transpose(0, 1)
        self.state = torch.einsum("mnoi,boin->boim", gate, self.state)

    def ry(self, theta: torch.Tensor, is_dagger: bool = False):
        """
        Apply Rotation-Y gate to the state vector.

        Arguments
        ---------
            :theta: torch.Tensor, shape: (out_dim, in_dim)
            :is_dagger: bool, default: False
        """
        gate = TorchGates.ry_gate(theta, dtype=self.dtype)
        if is_dagger:
            gate = torch.conj_physical(gate).transpose(0, 1)
        self.state = torch.einsum("mnoi,boin->boim", gate, self.state)

    def rz(self, theta: torch.Tensor, is_dagger: bool = False):
        """
        Apply Rotation-Z gate to the state vector.

        Arguments
        ---------
            :theta: torch.Tensor, shape: (out_dim, in_dim)
            :is_dagger: bool, default: False
        """
        gate = TorchGates.rz_gate(theta, dtype=self.dtype)
        if is_dagger:
            gate = torch.conj_physical(gate).transpose(0, 1)
        self.state = torch.einsum("mnoi,boin->boim", gate, self.state)


class DQStateVector:
    """
    2-qubit state vector.

    DQStateVector.state: torch.Tensor, shape: (batch_size, out_dim, in_dim, 4)
    """

    state: torch.Tensor

    def __init__(
        self,
        batch_size: int,
        out_dim: int,
        in_dim: int,
        device="cpu",
        dtype=torch.complex64,
    ):
        self.device = device
        self.batch_size = batch_size
        self.out_dim = out_dim
        self.in_dim = in_dim
        self.dtype = dtype
        self.state = torch.zeros(
            batch_size, out_dim, in_dim, 4, dtype=dtype, device=self.device
        )
        self.state[:, :, :, 0] = 1.0

    def measure_z(self, target: int = 0) -> torch.Tensor:
        """
        Measure the state vector in the Z basis.

        return: torch.Tensor, shape: (batch_size, out_dim, in_dim)
        """
        if target == 0:
            return (
                +self.state[:, :, :, 0].abs()
                - self.state[:, :, :, 1].abs()
                + self.state[:, :, :, 2].abs()
                - self.state[:, :, :, 3].abs()
            )
        else:
            return (
                +self.state[:, :, :, 0].abs()
                + self.state[:, :, :, 1].abs()
                - self.state[:, :, :, 2].abs()
                - self.state[:, :, :, 3].abs()
            )

    def cx(self, control: int):
        """
        Apply CX (CNOT) gate to the state vector.

        Arguments
        ---------
            :control: int
        """
        cx_gate = TorchGates.cx_gate(
            self.state.shape[1:3], control, self.device, dtype=self.dtype
        )
        self.state = torch.einsum("mnoi,boin->boim", cx_gate, self.state)

    def cz(self):
        """
        Apply CZ gate to the state vector.
        """
        cz_gate = TorchGates.cz_gate(
            self.state.shape[1:3], self.device, dtype=self.dtype
        )
        self.state = torch.einsum("mnoi,boin->boim", cz_gate, self.state)

    def apply_gate(self, gate: torch.Tensor, target: int = 0):
        """
        Apply a gate to the state vector.

        Arguments
        ---------
            :gate: torch.Tensor, shape: (4, 4, out_dim, in_dim)
        """
        if target == 0:
            gate = TorchGates.tensor_product(
                gate, TorchGates.identity_gate(self.state.shape[1:3])
            )
        else:
            gate = TorchGates.tensor_product(
                TorchGates.identity_gate(self.state.shape[1:3]), gate
            )
        self.state = torch.einsum("mnoi,boin->boim", gate, self.state)

    def apply_2gates(self, gate1: torch.Tensor, gate2: torch.Tensor):
        """
        Apply two gates to the state vector.

        Arguments
        ---------
            :gate1: torch.Tensor, shape: (4, 4, out_dim, in_dim)
            :gate2: torch.Tensor, shape: (4, 4, out_dim, in_dim)
        """
        gate = TorchGates.tensor_product(gate1, gate2)
        self.state = torch.einsum("mnoi,boin->boim", gate, self.state)

    def hh(self):
        """
        Apply Hadamard gate to the state vector.

        Arguments
        ---------
            :is_dagger: bool, default: False
        """
        h_gate = TorchGates.h_gate(self.state.shape[1:3], self.device, dtype=self.dtype)
        self.apply_2gates(h_gate, h_gate)



# MIT License
#
# Copyright (c) 2024 Huanqi Cao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Adoption of efficient-kan with implementation of conversion from QKAN to KAN.

Modified from [Blealtan/efficient-kan@GitHub 605b8c2](https://github.com/Blealtan/efficient-kan/tree/605b8c2ae24b8085bd11b96896a17eabe12f6736)
"""

import math
from typing import Union

import torch
import torch.nn.functional as F

from .qkan import QKAN


class KANLinear(torch.nn.Module):
    grid: torch.Tensor

    def __init__(
        self,
        in_features: int,
        out_features: int,
        grid_size: int = 5,
        spline_order: int = 3,
        scale_noise: float = 0.1,
        scale_base: float = 1.0,
        scale_spline: float = 1.0,
        enable_standalone_scale_spline: bool = True,
        base_activation=torch.nn.SiLU,
        grid_eps=0.02,
        grid_range=[-1, 1],
    ):
        super(KANLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.grid_size = grid_size
        self.spline_order = spline_order

        h = (grid_range[1] - grid_range[0]) / grid_size
        grid: torch.Tensor = (
            (
                torch.arange(-spline_order, grid_size + spline_order + 1) * h
                + grid_range[0]
            )
            .expand(in_features, -1)
            .contiguous()
        )
        self.register_buffer("grid", grid)

        self.base_weight = torch.nn.Parameter(torch.Tensor(out_features, in_features))
        self.spline_weight = torch.nn.Parameter(
            torch.Tensor(out_features, in_features, grid_size + spline_order)
        )
        if enable_standalone_scale_spline:
            self.spline_scaler = torch.nn.Parameter(
                torch.Tensor(out_features, in_features)
            )

        self.scale_noise = scale_noise
        self.scale_base = scale_base
        self.scale_spline = scale_spline
        self.enable_standalone_scale_spline = enable_standalone_scale_spline
        self.base_activation = base_activation()
        self.grid_eps = grid_eps

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.kaiming_uniform_(
            self.base_weight, a=math.sqrt(5) * self.scale_base
        )
        with torch.no_grad():
            noise = (
                (
                    torch.rand(self.grid_size + 1, self.in_features, self.out_features)
                    - 1 / 2
                )
                * self.scale_noise
                / self.grid_size
            )
            self.spline_weight.data.copy_(
                (self.scale_spline if not self.enable_standalone_scale_spline else 1.0)
                * self.curve2coeff(
                    self.grid.T[self.spline_order : -self.spline_order],
                    noise,
                )
            )
            if self.enable_standalone_scale_spline:
                # torch.nn.init.constant_(self.spline_scaler, self.scale_spline)
                torch.nn.init.kaiming_uniform_(
                    self.spline_scaler, a=math.sqrt(5) * self.scale_spline
                )

    def b_splines(self, x: torch.Tensor):
        """
        Compute the B-spline bases for the given input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_features).

        Returns:
            torch.Tensor: B-spline bases tensor of shape (batch_size, in_features, grid_size + spline_order).
        """
        assert x.dim() == 2 and x.size(1) == self.in_features

        grid: torch.Tensor = (
            self.grid
        )  # (in_features, grid_size + 2 * spline_order + 1)
        x = x.unsqueeze(-1)
        bases = ((x >= grid[:, :-1]) & (x < grid[:, 1:])).to(x.dtype)
        for k in range(1, self.spline_order + 1):
            bases = (
                (x - grid[:, : -(k + 1)])
                / (grid[:, k:-1] - grid[:, : -(k + 1)])
                * bases[:, :, :-1]
            ) + (
                (grid[:, k + 1 :] - x)
                / (grid[:, k + 1 :] - grid[:, 1:(-k)])
                * bases[:, :, 1:]
            )

        assert bases.size() == (
            x.size(0),
            self.in_features,
            self.grid_size + self.spline_order,
        )
        return bases.contiguous()

    def curve2coeff(self, x: torch.Tensor, y: torch.Tensor):
        """
        Compute the coefficients of the curve that interpolates the given points.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_features).
            y (torch.Tensor): Output tensor of shape (batch_size, in_features, out_features).

        Returns:
            torch.Tensor: Coefficients tensor of shape (out_features, in_features, grid_size + spline_order).
        """
        assert x.dim() == 2 and x.size(1) == self.in_features
        assert y.size() == (x.size(0), self.in_features, self.out_features)

        A = self.b_splines(x).transpose(
            0, 1
        )  # (in_features, batch_size, grid_size + spline_order)
        B = y.transpose(0, 1)  # (in_features, batch_size, out_features)
        solution = torch.linalg.lstsq(
            A, B
        ).solution  # (in_features, grid_size + spline_order, out_features)
        result = solution.permute(
            2, 0, 1
        )  # (out_features, in_features, grid_size + spline_order)

        assert result.size() == (
            self.out_features,
            self.in_features,
            self.grid_size + self.spline_order,
        )
        return result.contiguous()

    @property
    def scaled_spline_weight(self):
        return self.spline_weight * (
            self.spline_scaler.unsqueeze(-1)
            if self.enable_standalone_scale_spline
            else 1.0
        )

    def forward(self, x: torch.Tensor):
        assert x.size(-1) == self.in_features
        original_shape = x.shape
        x = x.view(-1, self.in_features)

        base_output = F.linear(self.base_activation(x), self.base_weight)
        spline_output = F.linear(
            self.b_splines(x).view(x.size(0), -1),
            self.scaled_spline_weight.view(self.out_features, -1),
        )
        output = base_output + spline_output

        output = output.view(*original_shape[:-1], self.out_features)
        return output

    @torch.no_grad()
    def update_grid(self, x: torch.Tensor, margin=0.01):
        assert x.dim() == 2 and x.size(1) == self.in_features
        batch = x.size(0)

        splines = self.b_splines(x)  # (batch, in, coeff)
        splines = splines.permute(1, 0, 2)  # (in, batch, coeff)
        orig_coeff = self.scaled_spline_weight  # (out, in, coeff)
        orig_coeff = orig_coeff.permute(1, 2, 0)  # (in, coeff, out)
        unreduced_spline_output = torch.bmm(splines, orig_coeff)  # (in, batch, out)
        unreduced_spline_output = unreduced_spline_output.permute(
            1, 0, 2
        )  # (batch, in, out)

        # sort each channel individually to collect data distribution
        x_sorted = torch.sort(x, dim=0)[0]
        grid_adaptive = x_sorted[
            torch.linspace(
                0, batch - 1, self.grid_size + 1, dtype=torch.int64, device=x.device
            )
        ]

        uniform_step = (x_sorted[-1] - x_sorted[0] + 2 * margin) / self.grid_size
        grid_uniform = (
            torch.arange(
                self.grid_size + 1, dtype=torch.float32, device=x.device
            ).unsqueeze(1)
            * uniform_step
            + x_sorted[0]
            - margin
        )

        grid = self.grid_eps * grid_uniform + (1 - self.grid_eps) * grid_adaptive
        grid = torch.concatenate(
            [
                grid[:1]
                - uniform_step
                * torch.arange(self.spline_order, 0, -1, device=x.device).unsqueeze(1),
                grid,
                grid[-1:]
                + uniform_step
                * torch.arange(1, self.spline_order + 1, device=x.device).unsqueeze(1),
            ],
            dim=0,
        )

        self.grid.copy_(grid.T)
        self.spline_weight.data.copy_(self.curve2coeff(x, unreduced_spline_output))

    @torch.no_grad()
    def update_grid_from_qkan(self, x: torch.Tensor, y: torch.Tensor, margin=0.01):
        assert x.dim() == 2 and x.size(1) == self.in_features
        batch = x.size(0)

        # sort each channel individually to collect data distribution
        x_sorted = torch.sort(x, dim=0)[0]
        grid_adaptive = x_sorted[
            torch.linspace(
                0, batch - 1, self.grid_size + 1, dtype=torch.int64, device=x.device
            )
        ]

        uniform_step = (x_sorted[-1] - x_sorted[0] + 2 * margin) / self.grid_size
        grid_uniform = (
            torch.arange(
                self.grid_size + 1, dtype=torch.float32, device=x.device
            ).unsqueeze(1)
            * uniform_step
            + x_sorted[0]
            - margin
        )

        grid = self.grid_eps * grid_uniform + (1 - self.grid_eps) * grid_adaptive
        grid = torch.concatenate(
            [
                grid[:1]
                - uniform_step
                * torch.arange(self.spline_order, 0, -1, device=x.device).unsqueeze(1),
                grid,
                grid[-1:]
                + uniform_step
                * torch.arange(1, self.spline_order + 1, device=x.device).unsqueeze(1),
            ],
            dim=0,
        )

        self.grid.copy_(grid.T)
        self.spline_weight.data.copy_(self.curve2coeff(x, y))

    def fit_from_qkan(
        self, x0: torch.Tensor, y: torch.Tensor, max_iter: int = 200, tol: float = 1e-5
    ):
        """
        Fit KAN layer from QKAN with early stopping.

        Args
        ----
            x0: torch.Tensor
                Input tensor
            y: torch.Tensor
                Target tensor
            max_iter: int
                Maximum number of iterations, default: 200
            tol: float
                Tolerance for early stopping, default: 1e-5
        """
        self.train()
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-1)
        criterion = torch.nn.MSELoss()
        prev_loss = float("inf")
        for _ in range(max_iter):
            optimizer.zero_grad()
            x = self.forward(x0)
            loss = criterion(x, y)
            loss.backward()
            optimizer.step()
            # Early stopping if loss improvement is below tolerance
            if abs(prev_loss - loss.item()) < tol:
                break
            prev_loss = loss.item()

    def regularization_loss(self, regularize_activation=1.0, regularize_entropy=1.0):
        """
        Compute the regularization loss.

        This is a dumb simulation of the original L1 regularization as stated in the
        paper, since the original one requires computing absolutes and entropy from the
        expanded (batch, in_features, out_features) intermediate tensor, which is hidden
        behind the F.linear function if we want an memory efficient implementation.

        The L1 regularization is now computed as mean absolute value of the spline
        weights. The authors implementation also includes this term in addition to the
        sample-based regularization.
        """
        l1_fake = self.spline_weight.abs().mean(-1)
        regularization_loss_activation = l1_fake.sum()
        p = l1_fake / regularization_loss_activation
        regularization_loss_entropy = -torch.sum(p * p.log())
        return (
            regularize_activation * regularization_loss_activation
            + regularize_entropy * regularization_loss_entropy
        )


class KANModuleList(torch.nn.ModuleList):
    def __init__(self):
        super(KANModuleList, self).__init__()

    # make type hint for getitem method
    def __getitem__(self, idx) -> Union[KANLinear, "KANModuleList"]:  # type: ignore
        return super(KANModuleList, self).__getitem__(idx)


class KAN(torch.nn.Module):
    """
    KAN (Kolmogorov-Arnold Network) model.
    This is an efficient implementation of the KAN model, which is a neural network that uses
    B-spline as the learnable variational activation function.

    It can be used for regression tasks and can be initialized from a QKAN model.
    """

    def __init__(
        self,
        layers_hidden,
        grid_size=5,
        spline_order=3,
        scale_noise=0.1,
        scale_base=1.0,
        scale_spline=1.0,
        base_activation=torch.nn.SiLU,
        grid_eps=0.02,
        grid_range=[-1, 1],
        device="cpu",
        seed=0,
        **kwargs,
    ):
        super(KAN, self).__init__()

        torch.manual_seed(seed)

        self.grid_size = grid_size
        self.spline_order = spline_order

        self.layers = KANModuleList()
        for in_features, out_features in zip(layers_hidden, layers_hidden[1:]):
            self.layers.append(
                KANLinear(
                    in_features,
                    out_features,
                    grid_size=grid_size,
                    spline_order=spline_order,
                    scale_noise=scale_noise,
                    scale_base=scale_base,
                    scale_spline=scale_spline,
                    base_activation=base_activation,
                    grid_eps=grid_eps,
                    grid_range=grid_range,
                )
            )
        self.device = device
        self.to(device)

    def forward(self, x: torch.Tensor, update_grid=False):
        for layer in self.layers:
            if update_grid:
                layer.update_grid(x)
            x = layer(x)
        return x

    def regularization_loss(self, regularize_activation=1.0, regularize_entropy=1.0):
        return sum(
            layer.regularization_loss(regularize_activation, regularize_entropy)
            for layer in self.layers
        )

    def initialize_from_qkan(self, qkan: QKAN, x0: torch.Tensor, sampling: int = 100):
        """
        Initialize KAN from a QKAN.

        Args
        ----
            qkan: QKAN
            x0: torch.Tensor (batch, in_dim)
            sampling: int
        """
        assert len(self.layers) == len(qkan.layers), "Mismatched architecture"
        if qkan.is_map:
            raise RuntimeError("Cannot initialize from a QKAN with a map layer")

        for i, kan_layer, qkan_layer in zip(
            range(len(self.layers)), self.layers, qkan.layers
        ):
            if i == 0:
                x = x0
            else:
                ymin = torch.min(out.cpu().detach(), dim=0).values  # noqa: F821
                ymax = torch.max(out.cpu().detach(), dim=0).values  # noqa: F821
                x = torch.stack(
                    [
                        torch.linspace(
                            ymin[i],
                            ymax[i],
                            steps=sampling,
                            device=x0.device,
                        )
                        for i in range(qkan_layer.in_dim)
                    ]
                ).permute(1, 0)  # x.shape = (sampling, in_dim)
            y = qkan_layer.forward_no_sum(x).transpose(
                1, 2
            )  # y.shape = (sampling, in_dim, out_dim)
            with torch.no_grad():
                out: torch.Tensor = qkan_layer.forward(
                    x
                )  # out.shape = (batch, out_dim)
            kan_layer.update_grid_from_qkan(x, y)
            kan_layer.fit_from_qkan(x, out)



# Copyright (c) 2026, Jiun-Cheng Jiang. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Triton-fused kernels for QKAN quantum circuit simulation.

Implements pz_encoding, rpz_encoding, and real ansatz forward passes as fused
Triton kernels, avoiding materialization of intermediate complex state vectors.
"""

import torch
import triton  # type: ignore
import triton.language as tl  # type: ignore


@triton.jit
def _pz_encoding_kernel(
    # Pointers
    x_ptr,  # [Batch, In_Dim]
    theta_ptr,  # [Out_Dim, In_Dim, Reps+1, 2]
    pw_ptr,  # [Out_Dim, In_Dim, Reps]  (preacts weights)
    pb_ptr,  # [Out_Dim, In_Dim, Reps]  (preacts bias)
    out_ptr,  # [Batch, Out_Dim, In_Dim]
    # Shapes
    batch_size,
    in_dim,
    out_dim,
    reps,
    # Strides for x
    stride_x_b,
    stride_x_i,
    # Strides for theta
    stride_t_o,
    stride_t_i,
    stride_t_r,
    stride_t_p,
    # Strides for preacts weight
    stride_pw_o,
    stride_pw_i,
    stride_pw_r,
    # Strides for preacts bias
    stride_pb_o,
    stride_pb_i,
    stride_pb_r,
    # Strides for output
    stride_o_b,
    stride_o_o,
    stride_o_i,
    # Compile-time constants
    PREACTS_TRAINABLE: tl.constexpr,
    FAST_MEASURE: tl.constexpr,
):
    """
    Fused QKAN pz_encoding forward kernel.

    Each program instance computes one element of the output[batch, out, in].
    The quantum circuit:
        H |0> -> for l in reps: Rz(theta[l,0]) Ry(theta[l,1]) Rz(enc_x) -> Rz(theta[L,0]) Ry(theta[L,1]) -> measure Z
    """
    pid = tl.program_id(0)

    # Decompose linear PID -> (batch, out, in)
    idx_i = pid % in_dim
    tmp = pid // in_dim
    idx_o = tmp % out_dim
    idx_b = tmp // out_dim

    if idx_b >= batch_size:
        return

    # Load input x[b, i]
    x_val = tl.load(x_ptr + idx_b * stride_x_b + idx_i * stride_x_i)

    # Initialize state: H|0> = (1/sqrt(2))|0> + (1/sqrt(2))|1>
    # State is (r0 + j*i0)|0> + (r1 + j*i1)|1>
    INV_SQRT2: tl.constexpr = 0.7071067811865476
    r0 = INV_SQRT2
    i0 = 0.0
    r1 = INV_SQRT2
    i1 = 0.0

    # Base pointer for theta[o, i, :, :]
    theta_base = theta_ptr + idx_o * stride_t_o + idx_i * stride_t_i

    # Main circuit loop
    for layer in range(reps):
        # Load theta parameters for this layer
        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        t1 = tl.load(theta_base + layer * stride_t_r + 1 * stride_t_p)

        # --- Apply Rz(t0) ---
        # Rz(t) = diag(e^{-it/2}, e^{it/2})
        # e^{-it/2} = cos(t/2) - j*sin(t/2)
        # |0> component: (r0+ji0) * (c-js) = (r0*c+i0*s) + j(i0*c-r0*s)
        # |1> component: (r1+ji1) * (c+js) = (r1*c-i1*s) + j(i1*c+r1*s)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = r0 * c + i0 * s
        ni0 = i0 * c - r0 * s
        nr1 = r1 * c - i1 * s
        ni1 = i1 * c + r1 * s
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        # --- Apply Ry(t1) ---
        # Ry(t) = [[cos(t/2), -sin(t/2)], [sin(t/2), cos(t/2)]]
        a = t1 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = c * r0 - s * r1
        ni0 = c * i0 - s * i1
        nr1 = s * r0 + c * r1
        ni1 = s * i0 + c * i1
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        # --- Apply Rz(encoded_x) (data re-uploading) ---
        enc = x_val
        if PREACTS_TRAINABLE:
            w = tl.load(
                pw_ptr + idx_o * stride_pw_o + idx_i * stride_pw_i + layer * stride_pw_r
            )
            b = tl.load(
                pb_ptr + idx_o * stride_pb_o + idx_i * stride_pb_i + layer * stride_pb_r
            )
            enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = r0 * c + i0 * s
        ni0 = i0 * c - r0 * s
        nr1 = r1 * c - i1 * s
        ni1 = i1 * c + r1 * s
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

    # Final rotations: Rz(theta[reps, 0]), Ry(theta[reps, 1])
    t0 = tl.load(theta_base + reps * stride_t_r + 0 * stride_t_p)
    t1 = tl.load(theta_base + reps * stride_t_r + 1 * stride_t_p)

    # Rz(t0)
    a = t0 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)
    nr0 = r0 * c + i0 * s
    ni0 = i0 * c - r0 * s
    nr1 = r1 * c - i1 * s
    ni1 = i1 * c + r1 * s
    r0, i0, r1, i1 = nr0, ni0, nr1, ni1

    # Ry(t1)
    a = t1 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)
    nr0 = c * r0 - s * r1
    ni0 = c * i0 - s * i1
    nr1 = s * r0 + c * r1
    ni1 = s * i0 + c * i1
    r0, i0, r1, i1 = nr0, ni0, nr1, ni1

    # Measure Z expectation value
    if FAST_MEASURE:
        # Quantum-inspired: |alpha| - |beta|
        result = tl.sqrt(r0 * r0 + i0 * i0) - tl.sqrt(r1 * r1 + i1 * i1)
    else:
        # Standard quantum: |alpha|^2 - |beta|^2
        result = (r0 * r0 + i0 * i0) - (r1 * r1 + i1 * i1)

    # Store result
    tl.store(
        out_ptr + idx_b * stride_o_b + idx_o * stride_o_o + idx_i * stride_o_i,
        result,
    )


def triton_pz_forward(
    x: torch.Tensor,
    theta: torch.Tensor,
    preacts_w: torch.Tensor,
    preacts_b: torch.Tensor,
    preacts_trainable: bool,
    fast_measure: bool = True,
) -> torch.Tensor:
    """
    Launch the Triton pz_encoding kernel.

    Args:
        x: (batch, in_dim) float32 on CUDA
        theta: (out_dim, in_dim, reps+1, 2) float32 on CUDA, already expanded
        preacts_w: (out_dim, in_dim, reps) float32, or any tensor if not trainable
        preacts_b: (out_dim, in_dim, reps) float32, or any tensor if not trainable
        preacts_trainable: whether preacts are used
        fast_measure: True for |alpha|-|beta|, False for |alpha|^2-|beta|^2

    Returns:
        (batch, out_dim, in_dim) float32
    """
    batch, in_dim = x.shape
    out_dim = theta.shape[0]
    reps = theta.shape[2] - 1

    x = x.contiguous()
    theta = theta.contiguous()

    output = torch.empty(batch, out_dim, in_dim, device=x.device, dtype=x.dtype)

    grid = (batch * out_dim * in_dim,)

    if preacts_trainable:
        preacts_w = preacts_w.contiguous()
        preacts_b = preacts_b.contiguous()
        pw_strides = (preacts_w.stride(0), preacts_w.stride(1), preacts_w.stride(2))
        pb_strides = (preacts_b.stride(0), preacts_b.stride(1), preacts_b.stride(2))
    else:
        # Kernel won't access these (dead code), pass dummy strides
        pw_strides = (0, 0, 0)
        pb_strides = (0, 0, 0)

    _pz_encoding_kernel[grid](
        x,
        theta,
        preacts_w,
        preacts_b,
        output,
        batch,
        in_dim,
        out_dim,
        reps,
        x.stride(0),
        x.stride(1),
        theta.stride(0),
        theta.stride(1),
        theta.stride(2),
        theta.stride(3),
        *pw_strides,
        *pb_strides,
        output.stride(0),
        output.stride(1),
        output.stride(2),
        PREACTS_TRAINABLE=preacts_trainable,
        FAST_MEASURE=fast_measure,
    )

    return output


# ── rpz_encoding kernel ─────────────────────────────────────────────────────


@triton.jit
def _rpz_encoding_kernel(
    # Pointers
    x_ptr,  # [Batch, In_Dim]
    theta_ptr,  # [Out_Dim, In_Dim, Reps+1, 1]
    pw_ptr,  # [Out_Dim, In_Dim, Reps]  (preacts weights, always loaded)
    pb_ptr,  # [Out_Dim, In_Dim, Reps]  (preacts bias, always loaded)
    out_ptr,  # [Batch, Out_Dim, In_Dim]
    # Shapes
    batch_size,
    in_dim,
    out_dim,
    reps,
    # Strides for x
    stride_x_b,
    stride_x_i,
    # Strides for theta
    stride_t_o,
    stride_t_i,
    stride_t_r,
    stride_t_p,
    # Strides for preacts weight
    stride_pw_o,
    stride_pw_i,
    stride_pw_r,
    # Strides for preacts bias
    stride_pb_o,
    stride_pb_i,
    stride_pb_r,
    # Strides for output
    stride_o_b,
    stride_o_o,
    stride_o_i,
    # Compile-time constants
    FAST_MEASURE: tl.constexpr,
):
    """
    Fused rpz_encoding forward kernel.

    Circuit: H|0> -> [Ry(theta[l,0]) Rz(w*x+b)]×reps -> Ry(theta[L,0]) -> measure Z
    rpz always uses encoded_x = w*x + b (even when preacts_trainable=False, w=1).
    """
    pid = tl.program_id(0)

    idx_i = pid % in_dim
    tmp = pid // in_dim
    idx_o = tmp % out_dim
    idx_b = tmp // out_dim

    if idx_b >= batch_size:
        return

    x_val = tl.load(x_ptr + idx_b * stride_x_b + idx_i * stride_x_i)

    # H|0> = (1/sqrt(2))|0> + (1/sqrt(2))|1>
    INV_SQRT2: tl.constexpr = 0.7071067811865476
    r0 = INV_SQRT2
    i0 = 0.0
    r1 = INV_SQRT2
    i1 = 0.0

    theta_base = theta_ptr + idx_o * stride_t_o + idx_i * stride_t_i
    pw_base = pw_ptr + idx_o * stride_pw_o + idx_i * stride_pw_i
    pb_base = pb_ptr + idx_o * stride_pb_o + idx_i * stride_pb_i

    for layer in range(reps):
        # Ry(theta[l, 0])
        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = c * r0 - s * r1
        ni0 = c * i0 - s * i1
        nr1 = s * r0 + c * r1
        ni1 = s * i0 + c * i1
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        # Rz(w * x + b)  — always load preacts
        w = tl.load(pw_base + layer * stride_pw_r)
        b = tl.load(pb_base + layer * stride_pb_r)
        enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = r0 * c + i0 * s
        ni0 = i0 * c - r0 * s
        nr1 = r1 * c - i1 * s
        ni1 = i1 * c + r1 * s
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

    # Final Ry(theta[reps, 0])
    t0 = tl.load(theta_base + reps * stride_t_r + 0 * stride_t_p)
    a = t0 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)
    nr0 = c * r0 - s * r1
    ni0 = c * i0 - s * i1
    nr1 = s * r0 + c * r1
    ni1 = s * i0 + c * i1
    r0, i0, r1, i1 = nr0, ni0, nr1, ni1

    # Measure Z
    if FAST_MEASURE:
        result = tl.sqrt(r0 * r0 + i0 * i0) - tl.sqrt(r1 * r1 + i1 * i1)
    else:
        result = (r0 * r0 + i0 * i0) - (r1 * r1 + i1 * i1)

    tl.store(
        out_ptr + idx_b * stride_o_b + idx_o * stride_o_o + idx_i * stride_o_i,
        result,
    )


def triton_rpz_forward(
    x: torch.Tensor,
    theta: torch.Tensor,
    preacts_w: torch.Tensor,
    preacts_b: torch.Tensor,
    fast_measure: bool = True,
) -> torch.Tensor:
    """
    Launch the Triton rpz_encoding kernel.

    Args:
        x: (batch, in_dim) float32
        theta: (out_dim, in_dim, reps+1, 1) float32, already expanded
        preacts_w: (out_dim, in_dim, reps) float32, already expanded
        preacts_b: (out_dim, in_dim, reps) float32, already expanded
        fast_measure: measurement mode

    Returns:
        (batch, out_dim, in_dim) float32
    """
    batch, in_dim = x.shape
    out_dim = theta.shape[0]
    reps = theta.shape[2] - 1

    x = x.contiguous()
    theta = theta.contiguous()
    preacts_w = preacts_w.contiguous()
    preacts_b = preacts_b.contiguous()

    output = torch.empty(batch, out_dim, in_dim, device=x.device, dtype=x.dtype)
    grid = (batch * out_dim * in_dim,)

    _rpz_encoding_kernel[grid](
        x,
        theta,
        preacts_w,
        preacts_b,
        output,
        batch,
        in_dim,
        out_dim,
        reps,
        x.stride(0),
        x.stride(1),
        theta.stride(0),
        theta.stride(1),
        theta.stride(2),
        theta.stride(3),
        preacts_w.stride(0),
        preacts_w.stride(1),
        preacts_w.stride(2),
        preacts_b.stride(0),
        preacts_b.stride(1),
        preacts_b.stride(2),
        output.stride(0),
        output.stride(1),
        output.stride(2),
        FAST_MEASURE=fast_measure,
    )

    return output


@triton.jit
def _rpz_encoding_backward_kernel(
    # Pointers
    x_ptr,
    theta_ptr,
    pw_ptr,
    pb_ptr,
    grad_out_ptr,
    states_ptr,
    grad_theta_ptr,
    grad_x_ptr,
    grad_pw_ptr,
    grad_pb_ptr,
    # Shapes
    batch_size,
    in_dim,
    out_dim,
    reps,
    # Strides for x
    stride_x_b,
    stride_x_i,
    # Strides for theta
    stride_t_o,
    stride_t_i,
    stride_t_r,
    stride_t_p,
    # Strides for preacts weight
    stride_pw_o,
    stride_pw_i,
    stride_pw_r,
    # Strides for preacts bias
    stride_pb_o,
    stride_pb_i,
    stride_pb_r,
    # Strides for grad_output
    stride_go_b,
    stride_go_o,
    stride_go_i,
    # Strides for states buffer
    stride_s_n,
    stride_s_s,
    stride_s_c,
    # Strides for grad_theta
    stride_gt_o,
    stride_gt_i,
    stride_gt_r,
    stride_gt_p,
    # Strides for grad_x
    stride_gx_b,
    stride_gx_i,
    # Strides for grad_pw/pb
    stride_gpw_o,
    stride_gpw_i,
    stride_gpw_r,
    stride_gpb_o,
    stride_gpb_i,
    stride_gpb_r,
    # Compile-time constants
    FAST_MEASURE: tl.constexpr,
):
    """Backward kernel for rpz_encoding ansatz."""
    pid = tl.program_id(0)
    idx_i = pid % in_dim
    tmp = pid // in_dim
    idx_o = tmp % out_dim
    idx_b = tmp // out_dim

    if idx_b >= batch_size:
        return

    x_val = tl.load(x_ptr + idx_b * stride_x_b + idx_i * stride_x_i)
    theta_base = theta_ptr + idx_o * stride_t_o + idx_i * stride_t_i
    pw_base = pw_ptr + idx_o * stride_pw_o + idx_i * stride_pw_i
    pb_base = pb_ptr + idx_o * stride_pb_o + idx_i * stride_pb_i
    states_base = states_ptr + pid * stride_s_n

    # ── Phase 1: Forward recompute, saving states ──
    INV_SQRT2: tl.constexpr = 0.7071067811865476
    r0 = INV_SQRT2
    i0 = 0.0
    r1 = INV_SQRT2
    i1 = 0.0

    tl.store(states_base + 0 * stride_s_s + 0 * stride_s_c, r0)
    tl.store(states_base + 0 * stride_s_s + 1 * stride_s_c, i0)
    tl.store(states_base + 0 * stride_s_s + 2 * stride_s_c, r1)
    tl.store(states_base + 0 * stride_s_s + 3 * stride_s_c, i1)

    state_idx = 1
    for layer in range(reps):
        # Ry(theta[l,0])
        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = c * r0 - s * r1
        ni0 = c * i0 - s * i1
        nr1 = s * r0 + c * r1
        ni1 = s * i0 + c * i1
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
        tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
        tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
        tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
        state_idx += 1

        # Rz(w*x+b)
        w = tl.load(pw_base + layer * stride_pw_r)
        b = tl.load(pb_base + layer * stride_pb_r)
        enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = r0 * c + i0 * s
        ni0 = i0 * c - r0 * s
        nr1 = r1 * c - i1 * s
        ni1 = i1 * c + r1 * s
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
        tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
        tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
        tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
        state_idx += 1

    # Final Ry(theta[reps,0])
    t0 = tl.load(theta_base + reps * stride_t_r + 0 * stride_t_p)
    a = t0 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)
    nr0 = c * r0 - s * r1
    ni0 = c * i0 - s * i1
    nr1 = s * r0 + c * r1
    ni1 = s * i0 + c * i1
    r0, i0, r1, i1 = nr0, ni0, nr1, ni1

    # ── Phase 2: Measurement gradient ──
    go = tl.load(
        grad_out_ptr + idx_b * stride_go_b + idx_o * stride_go_o + idx_i * stride_go_i
    )

    if FAST_MEASURE:
        alpha_norm = tl.sqrt(r0 * r0 + i0 * i0)
        beta_norm = tl.sqrt(r1 * r1 + i1 * i1)
        inv_alpha = tl.where(alpha_norm > 1e-30, 1.0 / alpha_norm, 0.0)
        inv_beta = tl.where(beta_norm > 1e-30, 1.0 / beta_norm, 0.0)
        ar0 = go * r0 * inv_alpha
        ai0 = go * i0 * inv_alpha
        ar1 = -go * r1 * inv_beta
        ai1 = -go * i1 * inv_beta
    else:
        ar0 = 2.0 * go * r0
        ai0 = 2.0 * go * i0
        ar1 = -2.0 * go * r1
        ai1 = -2.0 * go * i1

    # ── Phase 3: Backward sweep ──
    grad_x_local = 0.0
    gt_base = grad_theta_ptr + idx_o * stride_gt_o + idx_i * stride_gt_i

    # Backward through final Ry(theta[reps,0])
    state_idx -= 1
    sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
    si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
    sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
    si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

    t0 = tl.load(theta_base + reps * stride_t_r + 0 * stride_t_p)
    a = t0 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)

    grad_t0 = 0.5 * (
        ar0 * (-s * sr0 - c * sr1)
        + ai0 * (-s * si0 - c * si1)
        + ar1 * (c * sr0 - s * sr1)
        + ai1 * (c * si0 - s * si1)
    )
    tl.atomic_add(gt_base + reps * stride_gt_r + 0 * stride_gt_p, grad_t0)

    nar0 = c * ar0 + s * ar1
    nai0 = c * ai0 + s * ai1
    nar1 = -s * ar0 + c * ar1
    nai1 = -s * ai0 + c * ai1
    ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

    # Backward through layers in reverse
    for layer in range(reps - 1, -1, -1):
        # Backward through Rz(enc)
        state_idx -= 1
        sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
        si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
        sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
        si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

        w = tl.load(pw_base + layer * stride_pw_r)
        b = tl.load(pb_base + layer * stride_pb_r)
        enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)

        grad_enc = 0.5 * (
            -s * (ar0 * sr0 + ai0 * si0 + ar1 * sr1 + ai1 * si1)
            + c * (ar0 * si0 - ai0 * sr0 - ar1 * si1 + ai1 * sr1)
        )

        tl.atomic_add(
            grad_pw_ptr
            + idx_o * stride_gpw_o
            + idx_i * stride_gpw_i
            + layer * stride_gpw_r,
            grad_enc * x_val,
        )
        tl.atomic_add(
            grad_pb_ptr
            + idx_o * stride_gpb_o
            + idx_i * stride_gpb_i
            + layer * stride_gpb_r,
            grad_enc,
        )
        grad_x_local += grad_enc * w

        nar0 = c * ar0 - s * ai0
        nai0 = s * ar0 + c * ai0
        nar1 = c * ar1 + s * ai1
        nai1 = -s * ar1 + c * ai1
        ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

        # Backward through Ry(theta[l,0])
        state_idx -= 1
        sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
        si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
        sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
        si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)

        grad_t0 = 0.5 * (
            ar0 * (-s * sr0 - c * sr1)
            + ai0 * (-s * si0 - c * si1)
            + ar1 * (c * sr0 - s * sr1)
            + ai1 * (c * si0 - s * si1)
        )
        tl.atomic_add(gt_base + layer * stride_gt_r + 0 * stride_gt_p, grad_t0)

        nar0 = c * ar0 + s * ar1
        nai0 = c * ai0 + s * ai1
        nar1 = -s * ar0 + c * ar1
        nai1 = -s * ai0 + c * ai1
        ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

    tl.atomic_add(grad_x_ptr + idx_b * stride_gx_b + idx_i * stride_gx_i, grad_x_local)


def triton_rpz_backward(x, theta, pw, pb, grad_output, fast_measure):
    """Launch rpz_encoding backward kernel. Returns (grad_x, grad_theta, grad_pw, grad_pb)."""
    batch, in_dim = x.shape
    out_dim = theta.shape[0]
    reps = theta.shape[2] - 1

    x = x.contiguous()
    theta = theta.contiguous()
    pw = pw.contiguous()
    pb = pb.contiguous()
    grad_output = grad_output.contiguous()

    n_states = 2 * reps + 2  # H state + 2 per layer (after Ry, Rz)
    N = batch * out_dim * in_dim
    states = torch.empty(N, n_states, 4, device=x.device, dtype=x.dtype)

    grad_theta = torch.zeros_like(theta)
    grad_x = torch.zeros(batch, in_dim, device=x.device, dtype=x.dtype)
    grad_pw = torch.zeros_like(pw)
    grad_pb = torch.zeros_like(pb)

    grid = (N,)
    _rpz_encoding_backward_kernel[grid](
        x,
        theta,
        pw,
        pb,
        grad_output,
        states,
        grad_theta,
        grad_x,
        grad_pw,
        grad_pb,
        batch,
        in_dim,
        out_dim,
        reps,
        x.stride(0),
        x.stride(1),
        theta.stride(0),
        theta.stride(1),
        theta.stride(2),
        theta.stride(3),
        pw.stride(0),
        pw.stride(1),
        pw.stride(2),
        pb.stride(0),
        pb.stride(1),
        pb.stride(2),
        grad_output.stride(0),
        grad_output.stride(1),
        grad_output.stride(2),
        states.stride(0),
        states.stride(1),
        states.stride(2),
        grad_theta.stride(0),
        grad_theta.stride(1),
        grad_theta.stride(2),
        grad_theta.stride(3),
        grad_x.stride(0),
        grad_x.stride(1),
        grad_pw.stride(0),
        grad_pw.stride(1),
        grad_pw.stride(2),
        grad_pb.stride(0),
        grad_pb.stride(1),
        grad_pb.stride(2),
        FAST_MEASURE=fast_measure,
    )

    return grad_x, grad_theta, grad_pw, grad_pb


# ── real ansatz kernel ───────────────────────────────────────────────────────


@triton.jit
def _real_encoding_kernel(
    # Pointers
    x_ptr,  # [Batch, In_Dim]
    theta_ptr,  # [Out_Dim, In_Dim, Reps, 1]  (no final layer)
    pw_ptr,  # [Out_Dim, In_Dim, Reps]  (preacts weights)
    pb_ptr,  # [Out_Dim, In_Dim, Reps]  (preacts bias)
    out_ptr,  # [Batch, Out_Dim, In_Dim]
    # Shapes
    batch_size,
    in_dim,
    out_dim,
    reps,
    # Strides for x
    stride_x_b,
    stride_x_i,
    # Strides for theta
    stride_t_o,
    stride_t_i,
    stride_t_r,
    stride_t_p,
    # Strides for preacts weight
    stride_pw_o,
    stride_pw_i,
    stride_pw_r,
    # Strides for preacts bias
    stride_pb_o,
    stride_pb_i,
    stride_pb_r,
    # Strides for output
    stride_o_b,
    stride_o_o,
    stride_o_i,
    # Compile-time constants
    PREACTS_TRAINABLE: tl.constexpr,
    FAST_MEASURE: tl.constexpr,
):
    """
    Fused real ansatz forward kernel.

    Circuit: H|0> -> [X, Ry(theta[l,0]), Z, Ry(enc_x)]×reps -> measure Z
    No final rotation. Data encoding uses Ry (not Rz).
    """
    pid = tl.program_id(0)

    idx_i = pid % in_dim
    tmp = pid // in_dim
    idx_o = tmp % out_dim
    idx_b = tmp // out_dim

    if idx_b >= batch_size:
        return

    x_val = tl.load(x_ptr + idx_b * stride_x_b + idx_i * stride_x_i)

    # H|0>
    INV_SQRT2: tl.constexpr = 0.7071067811865476
    r0 = INV_SQRT2
    i0 = 0.0
    r1 = INV_SQRT2
    i1 = 0.0

    theta_base = theta_ptr + idx_o * stride_t_o + idx_i * stride_t_i

    for layer in range(reps):
        # X gate: swap |0> <-> |1>
        r0, i0, r1, i1 = r1, i1, r0, i0

        # Ry(theta[l, 0])
        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = c * r0 - s * r1
        ni0 = c * i0 - s * i1
        nr1 = s * r0 + c * r1
        ni1 = s * i0 + c * i1
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        # Z gate: negate |1> component
        r1 = -r1
        i1 = -i1

        # Ry(enc_x) — data encoding with Ry
        enc = x_val
        if PREACTS_TRAINABLE:
            w = tl.load(
                pw_ptr + idx_o * stride_pw_o + idx_i * stride_pw_i + layer * stride_pw_r
            )
            b = tl.load(
                pb_ptr + idx_o * stride_pb_o + idx_i * stride_pb_i + layer * stride_pb_r
            )
            enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = c * r0 - s * r1
        ni0 = c * i0 - s * i1
        nr1 = s * r0 + c * r1
        ni1 = s * i0 + c * i1
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

    # No final rotation — go straight to measurement

    # Measure Z
    if FAST_MEASURE:
        result = tl.sqrt(r0 * r0 + i0 * i0) - tl.sqrt(r1 * r1 + i1 * i1)
    else:
        result = (r0 * r0 + i0 * i0) - (r1 * r1 + i1 * i1)

    tl.store(
        out_ptr + idx_b * stride_o_b + idx_o * stride_o_o + idx_i * stride_o_i,
        result,
    )


# ── Backward kernels ───────────────────────────────────────────────────────


@triton.jit
def _pz_encoding_backward_kernel(
    # Pointers
    x_ptr,  # [Batch, In_Dim]
    theta_ptr,  # [Out_Dim, In_Dim, Reps+1, 2]
    pw_ptr,  # [Out_Dim, In_Dim, Reps]
    pb_ptr,  # [Out_Dim, In_Dim, Reps]
    grad_out_ptr,  # [Batch, Out_Dim, In_Dim]
    states_ptr,  # [batch*out*in, n_states, 4]
    grad_theta_ptr,  # [Out_Dim, In_Dim, Reps+1, 2]
    grad_x_ptr,  # [Batch, In_Dim]
    grad_pw_ptr,  # [Out_Dim, In_Dim, Reps]
    grad_pb_ptr,  # [Out_Dim, In_Dim, Reps]
    # Shapes
    batch_size,
    in_dim,
    out_dim,
    reps,
    # Strides for x
    stride_x_b,
    stride_x_i,
    # Strides for theta
    stride_t_o,
    stride_t_i,
    stride_t_r,
    stride_t_p,
    # Strides for preacts weight
    stride_pw_o,
    stride_pw_i,
    stride_pw_r,
    # Strides for preacts bias
    stride_pb_o,
    stride_pb_i,
    stride_pb_r,
    # Strides for grad_output
    stride_go_b,
    stride_go_o,
    stride_go_i,
    # Strides for states buffer [N, n_states, 4]
    stride_s_n,
    stride_s_s,
    stride_s_c,
    # Strides for grad_theta
    stride_gt_o,
    stride_gt_i,
    stride_gt_r,
    stride_gt_p,
    # Strides for grad_x
    stride_gx_b,
    stride_gx_i,
    # Strides for grad_pw/pb
    stride_gpw_o,
    stride_gpw_i,
    stride_gpw_r,
    stride_gpb_o,
    stride_gpb_i,
    stride_gpb_r,
    # Compile-time constants
    PREACTS_TRAINABLE: tl.constexpr,
    FAST_MEASURE: tl.constexpr,
):
    """Backward kernel for pz_encoding ansatz."""
    pid = tl.program_id(0)
    idx_i = pid % in_dim
    tmp = pid // in_dim
    idx_o = tmp % out_dim
    idx_b = tmp // out_dim

    if idx_b >= batch_size:
        return

    x_val = tl.load(x_ptr + idx_b * stride_x_b + idx_i * stride_x_i)
    theta_base = theta_ptr + idx_o * stride_t_o + idx_i * stride_t_i
    states_base = states_ptr + pid * stride_s_n

    # ── Phase 1: Forward recompute, saving states ──
    INV_SQRT2: tl.constexpr = 0.7071067811865476
    r0 = INV_SQRT2
    i0 = 0.0
    r1 = INV_SQRT2
    i1 = 0.0

    # Save initial state (state index 0)
    tl.store(states_base + 0 * stride_s_s + 0 * stride_s_c, r0)
    tl.store(states_base + 0 * stride_s_s + 1 * stride_s_c, i0)
    tl.store(states_base + 0 * stride_s_s + 2 * stride_s_c, r1)
    tl.store(states_base + 0 * stride_s_s + 3 * stride_s_c, i1)

    state_idx = 1
    for layer in range(reps):
        # Rz(t0)
        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = r0 * c + i0 * s
        ni0 = i0 * c - r0 * s
        nr1 = r1 * c - i1 * s
        ni1 = i1 * c + r1 * s
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
        tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
        tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
        tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
        state_idx += 1

        # Ry(t1)
        t1 = tl.load(theta_base + layer * stride_t_r + 1 * stride_t_p)
        a = t1 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = c * r0 - s * r1
        ni0 = c * i0 - s * i1
        nr1 = s * r0 + c * r1
        ni1 = s * i0 + c * i1
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
        tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
        tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
        tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
        state_idx += 1

        # Rz(enc)
        enc = x_val
        if PREACTS_TRAINABLE:
            w = tl.load(
                pw_ptr + idx_o * stride_pw_o + idx_i * stride_pw_i + layer * stride_pw_r
            )
            b = tl.load(
                pb_ptr + idx_o * stride_pb_o + idx_i * stride_pb_i + layer * stride_pb_r
            )
            enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = r0 * c + i0 * s
        ni0 = i0 * c - r0 * s
        nr1 = r1 * c - i1 * s
        ni1 = i1 * c + r1 * s
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
        tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
        tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
        tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
        state_idx += 1

    # Final Rz(t0)
    t0 = tl.load(theta_base + reps * stride_t_r + 0 * stride_t_p)
    a = t0 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)
    nr0 = r0 * c + i0 * s
    ni0 = i0 * c - r0 * s
    nr1 = r1 * c - i1 * s
    ni1 = i1 * c + r1 * s
    r0, i0, r1, i1 = nr0, ni0, nr1, ni1

    tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
    tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
    tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
    tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
    state_idx += 1

    # Final Ry(t1)
    t1 = tl.load(theta_base + reps * stride_t_r + 1 * stride_t_p)
    a = t1 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)
    nr0 = c * r0 - s * r1
    ni0 = c * i0 - s * i1
    nr1 = s * r0 + c * r1
    ni1 = s * i0 + c * i1
    r0, i0, r1, i1 = nr0, ni0, nr1, ni1
    # final state — no need to save

    # ── Phase 2: Measurement gradient ──
    go = tl.load(
        grad_out_ptr + idx_b * stride_go_b + idx_o * stride_go_o + idx_i * stride_go_i
    )

    if FAST_MEASURE:
        alpha_norm = tl.sqrt(r0 * r0 + i0 * i0)
        beta_norm = tl.sqrt(r1 * r1 + i1 * i1)
        # Avoid division by zero
        inv_alpha = tl.where(alpha_norm > 1e-30, 1.0 / alpha_norm, 0.0)
        inv_beta = tl.where(beta_norm > 1e-30, 1.0 / beta_norm, 0.0)
        ar0 = go * r0 * inv_alpha
        ai0 = go * i0 * inv_alpha
        ar1 = -go * r1 * inv_beta
        ai1 = -go * i1 * inv_beta
    else:
        ar0 = 2.0 * go * r0
        ai0 = 2.0 * go * i0
        ar1 = -2.0 * go * r1
        ai1 = -2.0 * go * i1

    # ── Phase 3: Backward sweep ──
    grad_x_local = 0.0
    # Total states saved = 3*reps + 2 (indices 0..state_idx-1 before final Ry)
    # We walk backward: final Ry, final Rz, then loop (Rz_enc, Ry, Rz) per layer reversed

    # -- Backward through final Ry(t1_final) --
    state_idx -= 1  # pre-Ry state
    sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
    si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
    sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
    si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

    t1 = tl.load(theta_base + reps * stride_t_r + 1 * stride_t_p)
    a = t1 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)

    # Param grad for Ry: v = 0.5*(-s*r0-c*r1, -s*i0-c*i1, c*r0-s*r1, c*i0-s*i1)
    grad_t1 = 0.5 * (
        ar0 * (-s * sr0 - c * sr1)
        + ai0 * (-s * si0 - c * si1)
        + ar1 * (c * sr0 - s * sr1)
        + ai1 * (c * si0 - s * si1)
    )

    gt_base = grad_theta_ptr + idx_o * stride_gt_o + idx_i * stride_gt_i
    tl.atomic_add(gt_base + reps * stride_gt_r + 1 * stride_gt_p, grad_t1)

    # Adjoint prop Ry^T: new_ar0=c*ar0+s*ar1, etc.
    nar0 = c * ar0 + s * ar1
    nai0 = c * ai0 + s * ai1
    nar1 = -s * ar0 + c * ar1
    nai1 = -s * ai0 + c * ai1
    ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

    # -- Backward through final Rz(t0_final) --
    state_idx -= 1
    sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
    si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
    sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
    si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

    t0 = tl.load(theta_base + reps * stride_t_r + 0 * stride_t_p)
    a = t0 * 0.5
    c = tl.cos(a)
    s = tl.sin(a)

    # Param grad for Rz
    grad_t0 = 0.5 * (
        -s * (ar0 * sr0 + ai0 * si0 + ar1 * sr1 + ai1 * si1)
        + c * (ar0 * si0 - ai0 * sr0 - ar1 * si1 + ai1 * sr1)
    )
    tl.atomic_add(gt_base + reps * stride_gt_r + 0 * stride_gt_p, grad_t0)

    # Adjoint prop Rz^T
    nar0 = c * ar0 - s * ai0
    nai0 = s * ar0 + c * ai0
    nar1 = c * ar1 + s * ai1
    nai1 = -s * ar1 + c * ai1
    ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

    # -- Backward through layers in reverse --
    for layer in range(reps - 1, -1, -1):
        # Backward through Rz(enc)
        state_idx -= 1
        sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
        si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
        sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
        si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

        enc = x_val
        if PREACTS_TRAINABLE:
            w = tl.load(
                pw_ptr + idx_o * stride_pw_o + idx_i * stride_pw_i + layer * stride_pw_r
            )
            b = tl.load(
                pb_ptr + idx_o * stride_pb_o + idx_i * stride_pb_i + layer * stride_pb_r
            )
            enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)

        grad_enc = 0.5 * (
            -s * (ar0 * sr0 + ai0 * si0 + ar1 * sr1 + ai1 * si1)
            + c * (ar0 * si0 - ai0 * sr0 - ar1 * si1 + ai1 * sr1)
        )

        if PREACTS_TRAINABLE:
            tl.atomic_add(
                grad_pw_ptr
                + idx_o * stride_gpw_o
                + idx_i * stride_gpw_i
                + layer * stride_gpw_r,
                grad_enc * x_val,
            )
            tl.atomic_add(
                grad_pb_ptr
                + idx_o * stride_gpb_o
                + idx_i * stride_gpb_i
                + layer * stride_gpb_r,
                grad_enc,
            )
            grad_x_local += grad_enc * w
        else:
            grad_x_local += grad_enc

        # Adjoint prop Rz^T
        nar0 = c * ar0 - s * ai0
        nai0 = s * ar0 + c * ai0
        nar1 = c * ar1 + s * ai1
        nai1 = -s * ar1 + c * ai1
        ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

        # Backward through Ry(t1)
        state_idx -= 1
        sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
        si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
        sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
        si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

        t1 = tl.load(theta_base + layer * stride_t_r + 1 * stride_t_p)
        a = t1 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)

        grad_t1 = 0.5 * (
            ar0 * (-s * sr0 - c * sr1)
            + ai0 * (-s * si0 - c * si1)
            + ar1 * (c * sr0 - s * sr1)
            + ai1 * (c * si0 - s * si1)
        )
        tl.atomic_add(gt_base + layer * stride_gt_r + 1 * stride_gt_p, grad_t1)

        nar0 = c * ar0 + s * ar1
        nai0 = c * ai0 + s * ai1
        nar1 = -s * ar0 + c * ar1
        nai1 = -s * ai0 + c * ai1
        ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

        # Backward through Rz(t0)
        state_idx -= 1
        sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
        si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
        sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
        si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)

        grad_t0 = 0.5 * (
            -s * (ar0 * sr0 + ai0 * si0 + ar1 * sr1 + ai1 * si1)
            + c * (ar0 * si0 - ai0 * sr0 - ar1 * si1 + ai1 * sr1)
        )
        tl.atomic_add(gt_base + layer * stride_gt_r + 0 * stride_gt_p, grad_t0)

        nar0 = c * ar0 - s * ai0
        nai0 = s * ar0 + c * ai0
        nar1 = c * ar1 + s * ai1
        nai1 = -s * ar1 + c * ai1
        ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

    # Accumulate grad_x across out_dim
    tl.atomic_add(grad_x_ptr + idx_b * stride_gx_b + idx_i * stride_gx_i, grad_x_local)


def triton_pz_backward(x, theta, pw, pb, grad_output, preacts_trainable, fast_measure):
    """Launch pz_encoding backward kernel. Returns (grad_x, grad_theta, grad_pw, grad_pb)."""
    batch, in_dim = x.shape
    out_dim = theta.shape[0]
    reps = theta.shape[2] - 1

    x = x.contiguous()
    theta = theta.contiguous()
    grad_output = grad_output.contiguous()

    n_states = (
        3 * reps + 3
    )  # H state + 3 per layer (after Rz, Ry, Rz_enc) + after final Rz
    N = batch * out_dim * in_dim
    states = torch.empty(N, n_states, 4, device=x.device, dtype=x.dtype)

    grad_theta = torch.zeros_like(theta)
    grad_x = torch.zeros(batch, in_dim, device=x.device, dtype=x.dtype)

    if preacts_trainable:
        pw = pw.contiguous()
        pb = pb.contiguous()
        grad_pw = torch.zeros_like(pw)
        grad_pb = torch.zeros_like(pb)
        pw_strides = (pw.stride(0), pw.stride(1), pw.stride(2))
        pb_strides = (pb.stride(0), pb.stride(1), pb.stride(2))
        gpw_strides = (grad_pw.stride(0), grad_pw.stride(1), grad_pw.stride(2))
        gpb_strides = (grad_pb.stride(0), grad_pb.stride(1), grad_pb.stride(2))
    else:
        grad_pw = torch.zeros(1, device=x.device)
        grad_pb = torch.zeros(1, device=x.device)
        pw_strides = (0, 0, 0)
        pb_strides = (0, 0, 0)
        gpw_strides = (0, 0, 0)
        gpb_strides = (0, 0, 0)

    grid = (N,)
    _pz_encoding_backward_kernel[grid](
        x,
        theta,
        pw,
        pb,
        grad_output,
        states,
        grad_theta,
        grad_x,
        grad_pw,
        grad_pb,
        batch,
        in_dim,
        out_dim,
        reps,
        x.stride(0),
        x.stride(1),
        theta.stride(0),
        theta.stride(1),
        theta.stride(2),
        theta.stride(3),
        *pw_strides,
        *pb_strides,
        grad_output.stride(0),
        grad_output.stride(1),
        grad_output.stride(2),
        states.stride(0),
        states.stride(1),
        states.stride(2),
        grad_theta.stride(0),
        grad_theta.stride(1),
        grad_theta.stride(2),
        grad_theta.stride(3),
        grad_x.stride(0),
        grad_x.stride(1),
        *gpw_strides,
        *gpb_strides,
        PREACTS_TRAINABLE=preacts_trainable,
        FAST_MEASURE=fast_measure,
    )

    return (
        grad_x,
        grad_theta,
        grad_pw if preacts_trainable else None,
        grad_pb if preacts_trainable else None,
    )


def triton_real_forward(
    x: torch.Tensor,
    theta: torch.Tensor,
    preacts_w: torch.Tensor,
    preacts_b: torch.Tensor,
    preacts_trainable: bool,
    fast_measure: bool = True,
) -> torch.Tensor:
    """
    Launch the Triton real ansatz kernel.

    Args:
        x: (batch, in_dim) float32
        theta: (out_dim, in_dim, reps, 1) float32, already expanded (no +1 layer)
        preacts_w: (out_dim, in_dim, reps) float32, or any tensor if not trainable
        preacts_b: (out_dim, in_dim, reps) float32, or any tensor if not trainable
        preacts_trainable: whether preacts are used
        fast_measure: measurement mode

    Returns:
        (batch, out_dim, in_dim) float32
    """
    batch, in_dim = x.shape
    out_dim = theta.shape[0]
    reps = theta.shape[2]  # No +1 for real ansatz

    x = x.contiguous()
    theta = theta.contiguous()

    output = torch.empty(batch, out_dim, in_dim, device=x.device, dtype=x.dtype)
    grid = (batch * out_dim * in_dim,)

    if preacts_trainable:
        preacts_w = preacts_w.contiguous()
        preacts_b = preacts_b.contiguous()
        pw_strides = (preacts_w.stride(0), preacts_w.stride(1), preacts_w.stride(2))
        pb_strides = (preacts_b.stride(0), preacts_b.stride(1), preacts_b.stride(2))
    else:
        pw_strides = (0, 0, 0)
        pb_strides = (0, 0, 0)

    _real_encoding_kernel[grid](
        x,
        theta,
        preacts_w,
        preacts_b,
        output,
        batch,
        in_dim,
        out_dim,
        reps,
        x.stride(0),
        x.stride(1),
        theta.stride(0),
        theta.stride(1),
        theta.stride(2),
        theta.stride(3),
        *pw_strides,
        *pb_strides,
        output.stride(0),
        output.stride(1),
        output.stride(2),
        PREACTS_TRAINABLE=preacts_trainable,
        FAST_MEASURE=fast_measure,
    )

    return output


@triton.jit
def _real_encoding_backward_kernel(
    # Pointers
    x_ptr,
    theta_ptr,
    pw_ptr,
    pb_ptr,
    grad_out_ptr,
    states_ptr,
    grad_theta_ptr,
    grad_x_ptr,
    grad_pw_ptr,
    grad_pb_ptr,
    # Shapes
    batch_size,
    in_dim,
    out_dim,
    reps,
    # Strides for x
    stride_x_b,
    stride_x_i,
    # Strides for theta
    stride_t_o,
    stride_t_i,
    stride_t_r,
    stride_t_p,
    # Strides for preacts weight
    stride_pw_o,
    stride_pw_i,
    stride_pw_r,
    # Strides for preacts bias
    stride_pb_o,
    stride_pb_i,
    stride_pb_r,
    # Strides for grad_output
    stride_go_b,
    stride_go_o,
    stride_go_i,
    # Strides for states buffer
    stride_s_n,
    stride_s_s,
    stride_s_c,
    # Strides for grad_theta
    stride_gt_o,
    stride_gt_i,
    stride_gt_r,
    stride_gt_p,
    # Strides for grad_x
    stride_gx_b,
    stride_gx_i,
    # Strides for grad_pw/pb
    stride_gpw_o,
    stride_gpw_i,
    stride_gpw_r,
    stride_gpb_o,
    stride_gpb_i,
    stride_gpb_r,
    # Compile-time constants
    PREACTS_TRAINABLE: tl.constexpr,
    FAST_MEASURE: tl.constexpr,
):
    """Backward kernel for real ansatz. Circuit: H|0> -> [X, Ry(theta), Z, Ry(enc)]×reps -> measure Z"""
    pid = tl.program_id(0)
    idx_i = pid % in_dim
    tmp = pid // in_dim
    idx_o = tmp % out_dim
    idx_b = tmp // out_dim

    if idx_b >= batch_size:
        return

    x_val = tl.load(x_ptr + idx_b * stride_x_b + idx_i * stride_x_i)
    theta_base = theta_ptr + idx_o * stride_t_o + idx_i * stride_t_i
    states_base = states_ptr + pid * stride_s_n

    # ── Phase 1: Forward recompute, saving states ──
    INV_SQRT2: tl.constexpr = 0.7071067811865476
    r0 = INV_SQRT2
    i0 = 0.0
    r1 = INV_SQRT2
    i1 = 0.0

    tl.store(states_base + 0 * stride_s_s + 0 * stride_s_c, r0)
    tl.store(states_base + 0 * stride_s_s + 1 * stride_s_c, i0)
    tl.store(states_base + 0 * stride_s_s + 2 * stride_s_c, r1)
    tl.store(states_base + 0 * stride_s_s + 3 * stride_s_c, i1)

    state_idx = 1
    for layer in range(reps):
        # X gate: swap |0> <-> |1>
        r0, i0, r1, i1 = r1, i1, r0, i0

        tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
        tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
        tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
        tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
        state_idx += 1

        # Ry(theta[l,0])
        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = c * r0 - s * r1
        ni0 = c * i0 - s * i1
        nr1 = s * r0 + c * r1
        ni1 = s * i0 + c * i1
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
        tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
        tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
        tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
        state_idx += 1

        # Z gate: negate |1>
        r1 = -r1
        i1 = -i1

        tl.store(states_base + state_idx * stride_s_s + 0 * stride_s_c, r0)
        tl.store(states_base + state_idx * stride_s_s + 1 * stride_s_c, i0)
        tl.store(states_base + state_idx * stride_s_s + 2 * stride_s_c, r1)
        tl.store(states_base + state_idx * stride_s_s + 3 * stride_s_c, i1)
        state_idx += 1

        # Ry(enc)
        enc = x_val
        if PREACTS_TRAINABLE:
            w = tl.load(
                pw_ptr + idx_o * stride_pw_o + idx_i * stride_pw_i + layer * stride_pw_r
            )
            b = tl.load(
                pb_ptr + idx_o * stride_pb_o + idx_i * stride_pb_i + layer * stride_pb_r
            )
            enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)
        nr0 = c * r0 - s * r1
        ni0 = c * i0 - s * i1
        nr1 = s * r0 + c * r1
        ni1 = s * i0 + c * i1
        r0, i0, r1, i1 = nr0, ni0, nr1, ni1

        # Don't save after Ry(enc): its output is the next iteration's
        # input (or the final state for measurement). This keeps state_idx
        # aligned so backward state_idx-1 points to the pre-gate state.

    # ── Phase 2: Measurement gradient ──
    go = tl.load(
        grad_out_ptr + idx_b * stride_go_b + idx_o * stride_go_o + idx_i * stride_go_i
    )

    if FAST_MEASURE:
        alpha_norm = tl.sqrt(r0 * r0 + i0 * i0)
        beta_norm = tl.sqrt(r1 * r1 + i1 * i1)
        inv_alpha = tl.where(alpha_norm > 1e-30, 1.0 / alpha_norm, 0.0)
        inv_beta = tl.where(beta_norm > 1e-30, 1.0 / beta_norm, 0.0)
        ar0 = go * r0 * inv_alpha
        ai0 = go * i0 * inv_alpha
        ar1 = -go * r1 * inv_beta
        ai1 = -go * i1 * inv_beta
    else:
        ar0 = 2.0 * go * r0
        ai0 = 2.0 * go * i0
        ar1 = -2.0 * go * r1
        ai1 = -2.0 * go * i1

    # ── Phase 3: Backward sweep ──
    grad_x_local = 0.0
    gt_base = grad_theta_ptr + idx_o * stride_gt_o + idx_i * stride_gt_i

    for layer in range(reps - 1, -1, -1):
        # Backward through Ry(enc) — pre-gate state = after Z = state[3l+3]
        state_idx -= 1
        sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
        si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
        sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
        si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

        enc = x_val
        if PREACTS_TRAINABLE:
            w = tl.load(
                pw_ptr + idx_o * stride_pw_o + idx_i * stride_pw_i + layer * stride_pw_r
            )
            b = tl.load(
                pb_ptr + idx_o * stride_pb_o + idx_i * stride_pb_i + layer * stride_pb_r
            )
            enc = w * x_val + b

        a = enc * 0.5
        c = tl.cos(a)
        s = tl.sin(a)

        grad_enc = 0.5 * (
            ar0 * (-s * sr0 - c * sr1)
            + ai0 * (-s * si0 - c * si1)
            + ar1 * (c * sr0 - s * sr1)
            + ai1 * (c * si0 - s * si1)
        )

        if PREACTS_TRAINABLE:
            tl.atomic_add(
                grad_pw_ptr
                + idx_o * stride_gpw_o
                + idx_i * stride_gpw_i
                + layer * stride_gpw_r,
                grad_enc * x_val,
            )
            tl.atomic_add(
                grad_pb_ptr
                + idx_o * stride_gpb_o
                + idx_i * stride_gpb_i
                + layer * stride_gpb_r,
                grad_enc,
            )
            grad_x_local += grad_enc * w
        else:
            grad_x_local += grad_enc

        # Adjoint prop Ry^T
        nar0 = c * ar0 + s * ar1
        nai0 = c * ai0 + s * ai1
        nar1 = -s * ar0 + c * ar1
        nai1 = -s * ai0 + c * ai1
        ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

        # Backward through Z gate: negate ar1, ai1 (no state needed)
        ar1 = -ar1
        ai1 = -ai1

        # Backward through Ry(theta[l,0]) — pre-gate state = after X = state[3l+1]
        state_idx -= 2  # skip over after-Ry(theta) state to reach after-X state
        sr0 = tl.load(states_base + state_idx * stride_s_s + 0 * stride_s_c)
        si0 = tl.load(states_base + state_idx * stride_s_s + 1 * stride_s_c)
        sr1 = tl.load(states_base + state_idx * stride_s_s + 2 * stride_s_c)
        si1 = tl.load(states_base + state_idx * stride_s_s + 3 * stride_s_c)

        t0 = tl.load(theta_base + layer * stride_t_r + 0 * stride_t_p)
        a = t0 * 0.5
        c = tl.cos(a)
        s = tl.sin(a)

        grad_t0 = 0.5 * (
            ar0 * (-s * sr0 - c * sr1)
            + ai0 * (-s * si0 - c * si1)
            + ar1 * (c * sr0 - s * sr1)
            + ai1 * (c * si0 - s * si1)
        )
        tl.atomic_add(gt_base + layer * stride_gt_r + 0 * stride_gt_p, grad_t0)

        nar0 = c * ar0 + s * ar1
        nai0 = c * ai0 + s * ai1
        nar1 = -s * ar0 + c * ar1
        nai1 = -s * ai0 + c * ai1
        ar0, ai0, ar1, ai1 = nar0, nai0, nar1, nai1

        # Backward through X gate: swap adjoint components (no state needed)
        ar0, ai0, ar1, ai1 = ar1, ai1, ar0, ai0

    tl.atomic_add(grad_x_ptr + idx_b * stride_gx_b + idx_i * stride_gx_i, grad_x_local)


def triton_real_backward(
    x, theta, pw, pb, grad_output, preacts_trainable, fast_measure
):
    """Launch real ansatz backward kernel. Returns (grad_x, grad_theta, grad_pw, grad_pb)."""
    batch, in_dim = x.shape
    out_dim = theta.shape[0]
    reps = theta.shape[2]  # No +1 for real ansatz

    x = x.contiguous()
    theta = theta.contiguous()
    grad_output = grad_output.contiguous()

    n_states = 3 * reps + 1  # H state + 3 per layer (after X, Ry_theta, Z)
    N = batch * out_dim * in_dim
    states = torch.empty(N, n_states, 4, device=x.device, dtype=x.dtype)

    grad_theta = torch.zeros_like(theta)
    grad_x = torch.zeros(batch, in_dim, device=x.device, dtype=x.dtype)

    if preacts_trainable:
        pw = pw.contiguous()
        pb = pb.contiguous()
        grad_pw = torch.zeros_like(pw)
        grad_pb = torch.zeros_like(pb)
        pw_strides = (pw.stride(0), pw.stride(1), pw.stride(2))
        pb_strides = (pb.stride(0), pb.stride(1), pb.stride(2))
        gpw_strides = (grad_pw.stride(0), grad_pw.stride(1), grad_pw.stride(2))
        gpb_strides = (grad_pb.stride(0), grad_pb.stride(1), grad_pb.stride(2))
    else:
        grad_pw = torch.zeros(1, device=x.device)
        grad_pb = torch.zeros(1, device=x.device)
        pw_strides = (0, 0, 0)
        pb_strides = (0, 0, 0)
        gpw_strides = (0, 0, 0)
        gpb_strides = (0, 0, 0)

    grid = (N,)
    _real_encoding_backward_kernel[grid](
        x,
        theta,
        pw,
        pb,
        grad_output,
        states,
        grad_theta,
        grad_x,
        grad_pw,
        grad_pb,
        batch,
        in_dim,
        out_dim,
        reps,
        x.stride(0),
        x.stride(1),
        theta.stride(0),
        theta.stride(1),
        theta.stride(2),
        theta.stride(3),
        *pw_strides,
        *pb_strides,
        grad_output.stride(0),
        grad_output.stride(1),
        grad_output.stride(2),
        states.stride(0),
        states.stride(1),
        states.stride(2),
        grad_theta.stride(0),
        grad_theta.stride(1),
        grad_theta.stride(2),
        grad_theta.stride(3),
        grad_x.stride(0),
        grad_x.stride(1),
        *gpw_strides,
        *gpb_strides,
        PREACTS_TRAINABLE=preacts_trainable,
        FAST_MEASURE=fast_measure,
    )

    return (
        grad_x,
        grad_theta,
        grad_pw if preacts_trainable else None,
        grad_pb if preacts_trainable else None,
    )



# Copyright (c) 2024, Jiun-Cheng Jiang. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
QKAN layer simulating solver

This module provides a solver for quantum neural networks using PyTorch, PennyLane,
Triton-accelerated kernels, or cuQuantum tensor network contraction.
"""

import math

import numpy as np
import torch

from .torch_qc import StateVector, TorchGates

# cuQuantum / opt_einsum availability for cutn_solver
try:
    from cuquantum.tensornet import contract_path as _cutn_contract_path  # type: ignore

    _CUTN_AVAILABLE = True
except ImportError:
    _CUTN_AVAILABLE = False

try:
    from opt_einsum import contract_path as _oe_contract_path  # type: ignore

    _OE_AVAILABLE = True
except ImportError:
    _OE_AVAILABLE = False

# Triton fused kernels for flash solver
try:
    from .fused_ops import (
        triton_pz_backward,
        triton_pz_forward,
        triton_real_backward,
        triton_real_forward,
        triton_rpz_backward,
        triton_rpz_forward,
    )

    _FLASH_AVAILABLE = True
except ImportError:
    _FLASH_AVAILABLE = False

_INV_SQRT2 = math.sqrt(2.0) / 2.0


def qml_solver(x: torch.Tensor, theta: torch.Tensor, reps: int, **kwargs):
    """
    Single-qubit data reuploading circuit using PennyLane.

    Args
    ----
        x : torch.Tensor
            shape: (batch_size, in_dim)
        theta : torch.Tensor
            shape: (reps, 2)
        reps : int
        qml_device : str
            default: "default.qubit"
    """
    import pennylane as qml  # type: ignore

    qml_device: str = kwargs.get("qml_device", "default.qubit")
    dev = qml.device(qml_device, wires=1)

    @qml.qnode(dev, interface="torch")
    def circuit(x: torch.Tensor, theta: torch.Tensor):
        """
        Args
        ----
            x : torch.Tensor
                shape: (batch_size, in_dim)
            theta : torch.Tensor
                shape: (reps, 2)
        """
        qml.RY(np.pi / 2, wires=0)
        for l in range(reps):
            qml.RZ(theta[l, 0], wires=0)
            qml.RY(theta[l, 1], wires=0)
            qml.RZ(x, wires=0)
        qml.RZ(theta[reps, 0], wires=0)
        qml.RY(theta[reps, 1], wires=0)
        return qml.expval(qml.PauliZ(0))

    return circuit(x, theta)


def torch_exact_solver(
    x: torch.Tensor,
    theta: torch.Tensor,
    preacts_weight: torch.Tensor,
    preacts_bias: torch.Tensor,
    reps: int,
    **kwargs,
) -> torch.Tensor:
    """
    Single-qubit data reuploading circuit.

    Args
    ----
        x : torch.Tensor
            shape: (batch_size, in_dim)
        theta : torch.Tensor
            shape: (*group, reps, 2)
        preacts_weight : torch.Tensor
            shape: (*group, reps)
        preacts_bias : torch.Tensor
            shape: (*group, reps)
        reps : int
        ansatz : str
            options: ["pz_encoding", "px_encoding"], default: "pz_encoding"
        n_group : int
            number of neurons in a group, default: in_dim of x

    Returns
    -------
        torch.Tensor
            shape: (batch_size, out_dim, in_dim)
    """
    batch, in_dim = x.shape
    device = x.device
    ansatz = kwargs.get("ansatz", "pz_encoding")
    # group = kwargs.get("group", in_dim)
    preacts_trainable = kwargs.get("preacts_trainable", False)
    fast_measure = kwargs.get("fast_measure", True)
    out_dim: int = kwargs.get("out_dim", in_dim)
    dtype = kwargs.get("dtype", torch.complex64)

    if len(theta.shape) != 4:
        theta = theta.unsqueeze(0)
    if theta.shape[1] != in_dim:
        repeat_out = out_dim
        repeat_in = in_dim // theta.shape[1] + 1
        theta = theta.repeat(repeat_out, repeat_in, 1, 1)[:, :in_dim, :, :]
    # rpz_encoding always needs encoded_x (with bias), even when preacts_trainable=False
    _needs_encoded_x = preacts_trainable or ansatz in ("rpz_encoding", "rpz")
    if _needs_encoded_x:
        if len(preacts_weight.shape) != 3:
            preacts_weight = preacts_weight.unsqueeze(0)
            preacts_bias = preacts_bias.unsqueeze(0)
        if preacts_weight.shape[1] != in_dim:
            repeat_out = out_dim
            repeat_in = in_dim // preacts_weight.shape[1] + 1
            preacts_weight = preacts_weight.repeat(repeat_out, repeat_in, 1)[
                :, :in_dim, :
            ]
            preacts_bias = preacts_bias.repeat(repeat_out, repeat_in, 1)[:, :in_dim, :]
        encoded_x = torch.einsum("oir,bi->boir", preacts_weight, x).add(preacts_bias)
        # encoded_x shape: (batch_size, out_dim, in_dim, reps)

    def _pz_encoding(theta: torch.Tensor):
        """
        Args
        ----
            theta : torch.Tensor
                shape: (*group, reps, 2)
        """
        psi = StateVector(
            x.shape[0],
            theta.shape[0],
            theta.shape[1],
            device=device,
            dtype=dtype,
        )  # psi.state: torch.Tensor, shape: (batch_size, out_dim, in_dim, 2)
        psi.h()
        if not preacts_trainable:
            rug = TorchGates.rz_gate(x, dtype=dtype)
        for l in range(reps):
            psi.rz(theta[:, :, l, 0])
            psi.ry(theta[:, :, l, 1])
            if not preacts_trainable:
                psi.state = torch.einsum("mnbi,boin->boim", rug, psi.state)
            else:
                psi.state = torch.einsum(
                    "mnboi,boin->boim",
                    TorchGates.rz_gate(encoded_x[:, :, :, l], dtype=dtype),
                    psi.state,
                )

        psi.rz(theta[:, :, reps, 0])
        psi.ry(theta[:, :, reps, 1])
        return psi.measure_z(fast_measure)  # shape: (batch_size, out_dim, in_dim)

    def _rpz_encoding(theta: torch.Tensor):
        """
        Args
        ----
            theta : torch.Tensor
                shape: (*group, reps, 2)
        """
        psi = StateVector(
            x.shape[0],
            theta.shape[0],
            theta.shape[1],
            device=device,
            dtype=dtype,
        )
        psi.h()
        for l in range(reps):
            psi.ry(theta[:, :, l, 0])
            psi.state = torch.einsum(
                "mnboi,boin->boim",
                TorchGates.rz_gate(encoded_x[:, :, :, l], dtype=dtype),
                psi.state,
            )
        psi.ry(theta[:, :, reps, 0])
        return psi.measure_z(fast_measure)  # shape: (batch_size, out_dim, in_dim)

    def _px_encoding(theta: torch.Tensor):
        """
        Args
        ----
            theta: torch.Tensor
                shape: (*group, reps, 1)
        """
        psi = StateVector(
            x.shape[0],
            theta.shape[0],
            theta.shape[1],
            device=device,
            dtype=dtype,
        )  # psi.state: torch.Tensor, shape: (batch_size * g, out_dim, n_group, 2)
        psi.h()
        for l in range(reps):
            psi.rz(theta[:, :, l, 0])
            psi.state = torch.einsum(
                "mnboi,boin->boim",
                TorchGates.rx_gate(
                    torch.acos(
                        # torch.sin(
                        encoded_x[:, :, :, l]
                        # )
                        # add sin to prevent input from exceeding pm 1
                    ),
                    dtype=dtype,
                ),
                psi.state,
            )
            """
            # complex extension implementation
            psi.state = torch.einsum(
                "mnboi,boin->boim",
                TorchGates.acrx_gate(
                    torch.einsum("oi,bi->boi", preacts_weight[:, :, l], x)
                ),
                psi.state,
            )
            """
        psi.rz(theta[:, :, reps, 0])
        return psi.measure_z(fast_measure)  # shape: (batch_size, out_dim, in_dim)

    def _real(theta: torch.Tensor):
        """
        Args
        ----
            theta: torch.Tensor
                shape: (*group, reps, 1)
        """
        psi = StateVector(
            x.shape[0],
            theta.shape[0],
            theta.shape[1],
            device=device,
            dtype=dtype,
        )  # psi.state: torch.Tensor, shape: (batch_size, out_dim, in_dim, 2)
        psi.h()
        if not preacts_trainable:
            rug = TorchGates.ry_gate(x, dtype=dtype)
        for l in range(reps):
            psi.x()
            # psi.z()
            psi.ry(theta[:, :, l, 0])
            psi.z()
            if not preacts_trainable:
                psi.state = torch.einsum("mnbi,boin->boim", rug, psi.state)
            else:
                psi.state = torch.einsum(
                    "mnboi,boin->boim",
                    TorchGates.ry_gate(encoded_x[:, :, :, l], dtype=dtype),
                    psi.state,
                )
        return psi.measure_z(fast_measure)  # shape: (batch_size, out_dim, in_dim)

    def _mix(theta: torch.Tensor):
        """
        Args
        ----
            theta: torch.Tensor
                shape: (*group, reps, 2)
        """
        psi = StateVector(
            x.shape[0],
            theta.shape[0],
            theta.shape[1],
            device=device,
            dtype=dtype,
        )  # psi.state: torch.Tensor, shape: (batch_size, out_dim, in_dim, 2)
        psi.h()
        if not preacts_trainable:
            rug_y = TorchGates.ry_gate(x, dtype=dtype)
        for l in range(reps):
            psi.rz(theta[:, :, l, 0])
            psi.rx(theta[:, :, l, 1])
            if not preacts_trainable:
                psi.state = torch.einsum("mnbi,boin->boim", rug_y, psi.state)
            else:
                psi.state = torch.einsum(
                    "mnboi,boin->boim",
                    TorchGates.ry_gate(encoded_x[:, :, :, l], dtype=dtype),
                    psi.state,
                )
        psi.rz(theta[:, :, reps, 0])
        psi.rx(theta[:, :, reps, 1])
        return psi.measure_z(fast_measure)  # shape: (batch_size, out_dim, in_dim)

    if ansatz == "pz_encoding" or ansatz == "pz":
        circuit = _pz_encoding
    elif ansatz == "rpz_encoding" or ansatz == "rpz":
        circuit = _rpz_encoding
    elif ansatz == "px_encoding" or ansatz == "px":
        circuit = _px_encoding
    elif ansatz == "real":
        circuit = _real
    elif ansatz == "mix":
        circuit = _mix
    elif callable(ansatz):
        circuit = ansatz
    else:
        raise NotImplementedError()
    x = circuit(theta)  # shape: (batch_size, out_dim, in_dim)
    return x


def _combined_xryz_gate(theta, dtype=torch.complex64):
    """
    Analytically compute X @ RY(theta) @ Z as a single 2x2 gate (real ansatz).

    X @ RY(θ) @ Z = [[sin(θ/2), -cos(θ/2)],
                      [cos(θ/2),  sin(θ/2)]]
    """
    cos = torch.cos(theta / 2)
    sin = torch.sin(theta / 2)
    return torch.stack(
        [
            torch.stack([sin, -cos]),
            torch.stack([cos, sin]),
        ]
    ).to(dtype)


def _combined_rz_ry_gate(alpha, beta, dtype=torch.complex64):
    """
    Fused gate for the pz ansatz sequence: first RZ(alpha), then RY(beta).

    Matrix product RY(β) @ RZ(α) (rightmost acts first on the state):

        [[cos(β/2)·e^{-iα/2}, -sin(β/2)·e^{+iα/2}],
         [sin(β/2)·e^{-iα/2},  cos(β/2)·e^{+iα/2}]]
    """
    cos = torch.cos(beta / 2)
    sin = torch.sin(beta / 2)
    exp_neg = torch.exp(-0.5j * alpha)
    exp_pos = torch.exp(0.5j * alpha)
    return torch.stack(
        [
            torch.stack([cos * exp_neg, -sin * exp_pos]),
            torch.stack([sin * exp_neg, cos * exp_pos]),
        ]
    ).to(dtype)


def _find_contraction_path(expression, operands):
    """Find optimal contraction path using cuQuantum or opt_einsum."""
    if _CUTN_AVAILABLE:
        path, _ = _cutn_contract_path(expression, *operands)
        return path
    if _OE_AVAILABLE:
        path, _ = _oe_contract_path(expression, *operands)
        return path
    return None


def _build_real_expression(reps, preacts_trainable):
    """
    Build einsum expression for the real-ansatz circuit.
    Circuit: |0> -> H -> [XRyZ(theta) -> RY(x)]^reps -> measure
    Operands per rep: 2 (fused gate + data encoding). No final gate.
    """
    chain = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    n_needed = 2 + 2 * reps
    if n_needed > 26:
        return None
    ci = 0
    q = chain[ci]
    ci += 1
    subs = [f"boi{q}"]
    q_new = chain[ci]
    ci += 1
    subs.append(f"{q_new}{q}")
    q = q_new
    for _ in range(reps):
        q_new = chain[ci]
        ci += 1
        subs.append(f"{q_new}{q}oi")
        q = q_new
        q_new = chain[ci]
        ci += 1
        subs.append(f"{q_new}{q}boi" if preacts_trainable else f"{q_new}{q}bi")
        q = q_new
    return ",".join(subs) + "->" + f"boi{q}"


def _build_pz_expression(reps, preacts_trainable):
    """
    Build einsum expression for the pz_encoding circuit.
    Circuit: |0> -> H -> [RzRy_fused(theta) -> RZ(x)]^reps -> RzRy_fused(theta_final) -> measure
    Operands per rep: 2 (fused gate + data encoding). +1 final gate.
    """
    chain = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    n_needed = 3 + 2 * reps  # H(2) + reps*2 + final(1)
    if n_needed > 26:
        return None
    ci = 0
    q = chain[ci]
    ci += 1
    subs = [f"boi{q}"]
    q_new = chain[ci]
    ci += 1
    subs.append(f"{q_new}{q}")
    q = q_new
    for _ in range(reps):
        q_new = chain[ci]
        ci += 1
        subs.append(f"{q_new}{q}oi")
        q = q_new
        q_new = chain[ci]
        ci += 1
        subs.append(f"{q_new}{q}boi" if preacts_trainable else f"{q_new}{q}bi")
        q = q_new
    # Final RzRy gate
    q_new = chain[ci]
    ci += 1
    subs.append(f"{q_new}{q}oi")
    q = q_new
    return ",".join(subs) + "->" + f"boi{q}"


def _build_rpz_expression(reps):
    """
    Build einsum expression for the rpz_encoding circuit.
    Circuit: |0> -> H -> [RY(theta) -> RZ(encoded_x)]^reps -> RY(theta_final) -> measure
    rpz always uses encoded_x so data gates are (batch, out, in).
    Operands per rep: 2 (RY + RZ_data). +1 final RY.
    """
    chain = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    n_needed = 3 + 2 * reps
    if n_needed > 26:
        return None
    ci = 0
    q = chain[ci]
    ci += 1
    subs = [f"boi{q}"]
    q_new = chain[ci]
    ci += 1
    subs.append(f"{q_new}{q}")
    q = q_new
    for _ in range(reps):
        q_new = chain[ci]
        ci += 1
        subs.append(f"{q_new}{q}oi")
        q = q_new
        q_new = chain[ci]
        ci += 1
        subs.append(f"{q_new}{q}boi")
        q = q_new
    # Final RY gate
    q_new = chain[ci]
    ci += 1
    subs.append(f"{q_new}{q}oi")
    q = q_new
    return ",".join(subs) + "->" + f"boi{q}"


# Cache for precompiled contraction plans: {(expression, shapes_tuple): plan}
_CUTN_PLAN_CACHE: dict = {}


def _precompile_plan(equation, operand_shapes):
    """
    Precompile a contraction plan: find the optimal path once and convert it
    into a list of pairwise einsum strings that can be executed without any
    string parsing in the hot path.

    Returns (steps, permute_str) or None if no path optimizer is available.
        steps: list of (idx1, idx2, einsum_str)
        permute_str: final transposition einsum or None
    """
    dummy_ops = [torch.empty(*s) for s in operand_shapes]
    path = _find_contraction_path(equation, dummy_ops)
    if path is None:
        return None

    input_str, output_str = equation.split("->")
    subscripts = input_str.split(",")
    final_indices = set(output_str)

    steps = []
    for i, j in path:
        idx1, idx2 = sorted((i, j))
        sub1, sub2 = subscripts[idx1], subscripts[idx2]

        remaining = [s for k, s in enumerate(subscripts) if k != idx1 and k != idx2]
        needed = set("".join(remaining)) | final_indices

        out_chars = [
            c for c in (sub1 + sub2) if c in (set(sub1) | set(sub2)) and c in needed
        ]
        out_sub = "".join(dict.fromkeys(out_chars))

        steps.append((idx1, idx2, f"{sub1},{sub2}->{out_sub}"))

        subscripts.pop(idx2)
        subscripts.pop(idx1)
        subscripts.append(out_sub)

    permute = f"{subscripts[0]}->{output_str}" if subscripts[0] != output_str else None
    return steps, permute


def _execute_plan(plan, operands):
    """Execute a precompiled contraction plan (hot path, no string parsing)."""
    steps, permute = plan
    ops = list(operands)
    for idx1, idx2, einsum_str in steps:
        new_op = torch.einsum(einsum_str, ops[idx1], ops[idx2])
        ops.pop(idx2)
        ops.pop(idx1)
        ops.append(new_op)
    if permute:
        return torch.einsum(permute, ops[0])
    return ops[0]


def _get_plan(expression, operands):
    """Get (or compute and cache) a contraction plan for the given expression."""
    key = (expression, tuple(op.shape for op in operands))
    if key not in _CUTN_PLAN_CACHE:
        _CUTN_PLAN_CACHE[key] = _precompile_plan(
            expression, [op.shape for op in operands]
        )
    return _CUTN_PLAN_CACHE[key]


def cutn_solver(
    x: torch.Tensor,
    theta: torch.Tensor,
    preacts_weight: torch.Tensor,
    preacts_bias: torch.Tensor,
    reps: int,
    **kwargs,
) -> torch.Tensor:
    """
    Tensor network contraction solver using optimal contraction paths.

    Expresses the entire quantum circuit as a single tensor network and
    contracts it using an optimal path from cuQuantum or opt_einsum.
    The contraction plan is precompiled and cached so repeated forward
    calls pay no path-finding overhead.

    Supports ``pz_encoding`` (``pz``), ``rpz_encoding`` (``rpz``), and ``real`` ansatzes.
    Falls back to torch_exact_solver for unsupported ansatzes or reps > 11.

    Args
    ----
        x : torch.Tensor
            shape: (batch_size, in_dim)
        theta : torch.Tensor
            shape: (*group, reps+1, n_params_per_gate)
        preacts_weight : torch.Tensor
            shape: (*group, reps)
        preacts_bias : torch.Tensor
            shape: (*group, reps)
        reps : int
        ansatz : str
            options: "pz_encoding", "pz", "rpz_encoding", "rpz", "real"
        preacts_trainable : bool
        fast_measure : bool
        out_dim : int
        dtype : torch.dtype

    Returns
    -------
        torch.Tensor
            shape: (batch_size, out_dim, in_dim)
    """
    batch, in_dim = x.shape
    device = x.device
    ansatz = kwargs.get("ansatz", "pz_encoding")
    preacts_trainable = kwargs.get("preacts_trainable", False)
    fast_measure = kwargs.get("fast_measure", True)
    out_dim: int = kwargs.get("out_dim", in_dim)
    dtype = kwargs.get("dtype", torch.complex64)

    _SUPPORTED = {"pz_encoding", "pz", "rpz_encoding", "rpz", "real"}
    if ansatz not in _SUPPORTED:
        return torch_exact_solver(
            x, theta, preacts_weight, preacts_bias, reps, **kwargs
        )

    # Build whole-circuit expression based on ansatz
    if ansatz in ("pz_encoding", "pz"):
        expression = _build_pz_expression(reps, preacts_trainable)
    elif ansatz in ("rpz_encoding", "rpz"):
        expression = _build_rpz_expression(reps)
    else:  # real
        expression = _build_real_expression(reps, preacts_trainable)

    if expression is None:  # reps too large for single-char indices
        return torch_exact_solver(
            x, theta, preacts_weight, preacts_bias, reps, **kwargs
        )

    # Broadcasting logic (same as torch_exact_solver)
    if len(theta.shape) != 4:
        theta = theta.unsqueeze(0)
    if theta.shape[1] != in_dim:
        repeat_out = out_dim
        repeat_in = in_dim // theta.shape[1] + 1
        theta = theta.repeat(repeat_out, repeat_in, 1, 1)[:, :in_dim, :, :]

    _needs_encoded_x = preacts_trainable or ansatz in ("rpz_encoding", "rpz")
    if _needs_encoded_x:
        if len(preacts_weight.shape) != 3:
            preacts_weight = preacts_weight.unsqueeze(0)
            preacts_bias = preacts_bias.unsqueeze(0)
        if preacts_weight.shape[1] != in_dim:
            repeat_out = out_dim
            repeat_in = in_dim // preacts_weight.shape[1] + 1
            preacts_weight = preacts_weight.repeat(repeat_out, repeat_in, 1)[
                :, :in_dim, :
            ]
            preacts_bias = preacts_bias.repeat(repeat_out, repeat_in, 1)[:, :in_dim, :]
        encoded_x = torch.einsum("oir,bi->boir", preacts_weight, x).add(preacts_bias)

    # Build 2x2 H gate
    inv_sqrt2 = torch.tensor(_INV_SQRT2, device=device, dtype=dtype)
    h_gate = torch.stack(
        [
            torch.stack([inv_sqrt2, inv_sqrt2]),
            torch.stack([inv_sqrt2, -inv_sqrt2]),
        ]
    )

    # -- Build initial state --
    psi = torch.zeros(batch, out_dim, in_dim, 2, dtype=dtype, device=device)
    psi[:, :, :, 0] = 1.0

    # -- Build operands based on ansatz --
    operands = [psi, h_gate]

    if ansatz in ("pz_encoding", "pz"):
        # Circuit: H -> [RZ(θ₀)·RY(θ₁) -> RZ(x)]^reps -> RZ(θ_f₀)·RY(θ_f₁)
        if not preacts_trainable:
            rz_data = TorchGates.rz_gate(x, dtype=dtype)  # (2, 2, batch, in_dim)
        for l in range(reps):
            fused_l = _combined_rz_ry_gate(
                theta[:, :, l, 0], theta[:, :, l, 1], dtype=dtype
            )
            operands.append(fused_l)
            if not preacts_trainable:
                operands.append(rz_data)
            else:
                operands.append(TorchGates.rz_gate(encoded_x[:, :, :, l], dtype=dtype))
        # Final fused gate
        operands.append(
            _combined_rz_ry_gate(
                theta[:, :, reps, 0], theta[:, :, reps, 1], dtype=dtype
            )
        )

    elif ansatz in ("rpz_encoding", "rpz"):
        # Circuit: H -> [RY(θ) -> RZ(encoded_x)]^reps -> RY(θ_final)
        for l in range(reps):
            operands.append(TorchGates.ry_gate(theta[:, :, l, 0], dtype=dtype))
            operands.append(TorchGates.rz_gate(encoded_x[:, :, :, l], dtype=dtype))
        # Final RY gate
        operands.append(TorchGates.ry_gate(theta[:, :, reps, 0], dtype=dtype))

    else:  # real
        # Circuit: H -> [X·RY(θ)·Z -> RY(x)]^reps
        if not preacts_trainable:
            ry_data = TorchGates.ry_gate(x, dtype=dtype)
        for l in range(reps):
            operands.append(_combined_xryz_gate(theta[:, :, l, 0], dtype=dtype))
            if not preacts_trainable:
                operands.append(ry_data)
            else:
                operands.append(TorchGates.ry_gate(encoded_x[:, :, :, l], dtype=dtype))

    # Get cached contraction plan (path computed only once per shape config)
    plan = _get_plan(expression, operands)

    if plan is not None:
        psi = _execute_plan(plan, operands)
    else:
        psi = torch.einsum(expression, *operands)

    # Measurement (Z basis)
    return (
        psi[:, :, :, 0].abs() - psi[:, :, :, 1].abs()
        if fast_measure
        else torch.square(psi[:, :, :, 0].abs()) - torch.square(psi[:, :, :, 1].abs())
    )


# ---------------------------------------------------------------------------
# Flash (Triton-accelerated) solver
# ---------------------------------------------------------------------------

_SUPPORTED_FLASH_ANSATZES = {"pz_encoding", "pz", "rpz_encoding", "rpz", "real"}


class _FlashFunction(torch.autograd.Function):
    """
    Custom autograd function: Triton forward and backward.

    Forward dispatches to the appropriate Triton kernel based on ansatz.
    Backward uses direct Triton kernels with forward recomputation.
    """

    @staticmethod
    def forward(
        ctx,
        x,
        theta,
        preacts_w,
        preacts_b,
        reps,
        fast_measure,
        preacts_trainable,
        out_dim,
        c_dtype,
        ansatz,
    ):
        ctx.save_for_backward(x, theta, preacts_w, preacts_b)
        ctx.reps = reps
        ctx.fast_measure = fast_measure
        ctx.preacts_trainable = preacts_trainable
        ctx.out_dim = out_dim
        ctx.c_dtype = c_dtype
        ctx.ansatz = ansatz

        if ansatz in ("pz_encoding", "pz"):
            return triton_pz_forward(
                x, theta, preacts_w, preacts_b, preacts_trainable, fast_measure
            )
        elif ansatz in ("rpz_encoding", "rpz"):
            return triton_rpz_forward(x, theta, preacts_w, preacts_b, fast_measure)
        elif ansatz == "real":
            return triton_real_forward(
                x, theta, preacts_w, preacts_b, preacts_trainable, fast_measure
            )
        else:
            raise ValueError(f"Unsupported ansatz for flash: {ansatz}")

    @staticmethod
    def backward(ctx, grad_output):
        x, theta, preacts_w, preacts_b = ctx.saved_tensors
        ansatz = ctx.ansatz

        if ansatz in ("pz_encoding", "pz"):
            grad_x, grad_theta, grad_pw, grad_pb = triton_pz_backward(
                x,
                theta,
                preacts_w,
                preacts_b,
                grad_output,
                ctx.preacts_trainable,
                ctx.fast_measure,
            )
        elif ansatz in ("rpz_encoding", "rpz"):
            grad_x, grad_theta, grad_pw, grad_pb = triton_rpz_backward(
                x,
                theta,
                preacts_w,
                preacts_b,
                grad_output,
                ctx.fast_measure,
            )
        elif ansatz == "real":
            grad_x, grad_theta, grad_pw, grad_pb = triton_real_backward(
                x,
                theta,
                preacts_w,
                preacts_b,
                grad_output,
                ctx.preacts_trainable,
                ctx.fast_measure,
            )
        else:
            raise ValueError(f"Unsupported ansatz for flash backward: {ansatz}")

        return (
            grad_x,
            grad_theta,
            grad_pw,
            grad_pb,
            None,  # reps
            None,  # fast_measure
            None,  # preacts_trainable
            None,  # out_dim
            None,  # c_dtype
            None,  # ansatz
        )


def flash_exact_solver(
    x: torch.Tensor,
    theta: torch.Tensor,
    preacts_weight: torch.Tensor,
    preacts_bias: torch.Tensor,
    reps: int,
    **kwargs,
) -> torch.Tensor:
    """
    Triton-accelerated exact solver. Drop-in replacement for torch_exact_solver.

    Uses fused Triton kernels for pz_encoding, rpz_encoding, and real ansatzes.
    Falls back to torch_exact_solver for unsupported ansatzes.

    Args:
        Same as torch_exact_solver.

    Returns:
        torch.Tensor, shape: (batch_size, out_dim, in_dim)
    """
    if not _FLASH_AVAILABLE:
        raise ImportError(
            "Triton fused kernels not available. Install triton to use flash solver."
        )

    ansatz = kwargs.get("ansatz", "pz_encoding")
    preacts_trainable = kwargs.get("preacts_trainable", False)
    fast_measure = kwargs.get("fast_measure", True)
    out_dim: int = kwargs.get("out_dim", x.shape[1])
    c_dtype = kwargs.get("dtype", torch.complex64)
    batch, in_dim = x.shape

    # Fallback for unsupported ansatzes
    if ansatz not in _SUPPORTED_FLASH_ANSATZES:
        return torch_exact_solver(
            x, theta, preacts_weight, preacts_bias, reps, **kwargs
        )

    # Broadcasting logic (mirrors torch_exact_solver)
    if len(theta.shape) != 4:
        theta = theta.unsqueeze(0)
    if theta.shape[1] != in_dim:
        repeat_out = out_dim
        repeat_in = in_dim // theta.shape[1] + 1
        theta = theta.repeat(repeat_out, repeat_in, 1, 1)[:, :in_dim, :, :]

    # rpz always needs encoded_x; others only when preacts_trainable
    _needs_encoded_x = preacts_trainable or ansatz in ("rpz_encoding", "rpz")
    if _needs_encoded_x:
        if len(preacts_weight.shape) != 3:
            preacts_weight = preacts_weight.unsqueeze(0)
            preacts_bias = preacts_bias.unsqueeze(0)
        if preacts_weight.shape[1] != in_dim:
            repeat_out = out_dim
            repeat_in = in_dim // preacts_weight.shape[1] + 1
            preacts_weight = preacts_weight.repeat(repeat_out, repeat_in, 1)[
                :, :in_dim, :
            ]
            preacts_bias = preacts_bias.repeat(repeat_out, repeat_in, 1)[:, :in_dim, :]

    # Check if gradients are needed (training)
    needs_grad = theta.requires_grad or x.requires_grad
    if _needs_encoded_x:
        needs_grad = (
            needs_grad or preacts_weight.requires_grad or preacts_bias.requires_grad
        )
    elif preacts_trainable:
        needs_grad = (
            needs_grad or preacts_weight.requires_grad or preacts_bias.requires_grad
        )

    if needs_grad:
        return _FlashFunction.apply(
            x,
            theta,
            preacts_weight,
            preacts_bias,
            reps,
            fast_measure,
            preacts_trainable,
            out_dim,
            c_dtype,
            ansatz,
        )
    else:
        if ansatz in ("pz_encoding", "pz"):
            return triton_pz_forward(
                x,
                theta,
                preacts_weight,
                preacts_bias,
                preacts_trainable,
                fast_measure,
            )
        elif ansatz in ("rpz_encoding", "rpz"):
            return triton_rpz_forward(
                x,
                theta,
                preacts_weight,
                preacts_bias,
                fast_measure,
            )
        elif ansatz == "real":
            return triton_real_forward(
                x,
                theta,
                preacts_weight,
                preacts_bias,
                preacts_trainable,
                fast_measure,
            )
        else:
            raise NotImplementedError



# MIT License

# Copyright (c) 2025 Andrej Karpathy

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Helpful info functions for QKAN.
Parts of codes are adapted from https://github.com/karpathy/nanochat.
"""

import os

from . import __version__


def print0(*s, **kwargs):
    ddp_rank = int(os.environ.get("RANK", 0))
    if ddp_rank == 0:
        print(*s, **kwargs)


def print_banner():
    # Cool DOS Rebel font ASCII banner made with https://manytools.org/hacker-tools/ascii-banner/
    banner = """
        ██████    █████   ████   █████████   ██████   █████
      ███░░░░███ ░░███   ███░   ███░░░░░███ ░░██████ ░░███ 
     ███    ░░███ ░███  ███    ░███    ░███  ░███░███ ░███ 
    ░███     ░███ ░███████     ░███████████  ░███░░███░███ 
    ░███   ██░███ ░███░░███    ░███░░░░░███  ░███ ░░██████ 
    ░░███ ░░████  ░███ ░░███   ░███    ░███  ░███  ░░█████ 
     ░░░██████░██ █████ ░░████ █████   █████ █████  ░░█████
       ░░░░░░ ░░ ░░░░░   ░░░░ ░░░░░   ░░░░░ ░░░░░    ░░░░░     
    """
    print0(banner)


def print_version():
    # print the version and the banner
    print0("=" * 60)
    print_banner()
    print0(f"QKAN version: {__version__}")
    print0("=" * 60)


def is_ddp():
    # TODO is there a proper way
    return int(os.environ.get("RANK", -1)) != -1


def get_dist_info():
    if is_ddp():
        assert all(var in os.environ for var in ["RANK", "LOCAL_RANK", "WORLD_SIZE"])
        ddp_rank = int(os.environ["RANK"])
        ddp_local_rank = int(os.environ["LOCAL_RANK"])
        ddp_world_size = int(os.environ["WORLD_SIZE"])
        return True, ddp_rank, ddp_local_rank, ddp_world_size
    else:
        return False, 0, 0, 1



# MIT License
#
# Copyright (c) 2024 Ziming Liu
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Create dataset for regression task.

Adapted from [KindXiaoming/pykan@GitHub 91a2f63](https://github.com/KindXiaoming/pykan/tree/91a2f633be2d435b081ef0ef52a7205c7e7bea9e)
"""

import numpy as np
import sympy
import torch

# sigmoid = sympy.Function('sigmoid')
# name: (torch implementation, sympy implementation)

# singularity protection functions
f_inv = lambda x, y_th: (
    (x_th := 1 / y_th),
    y_th / x_th * x * (torch.abs(x) < x_th)
    + torch.nan_to_num(1 / x) * (torch.abs(x) >= x_th),
)
f_inv2 = lambda x, y_th: (
    (x_th := 1 / y_th ** (1 / 2)),
    y_th * (torch.abs(x) < x_th) + torch.nan_to_num(1 / x**2) * (torch.abs(x) >= x_th),
)
f_inv3 = lambda x, y_th: (
    (x_th := 1 / y_th ** (1 / 3)),
    y_th / x_th * x * (torch.abs(x) < x_th)
    + torch.nan_to_num(1 / x**3) * (torch.abs(x) >= x_th),
)
f_inv4 = lambda x, y_th: (
    (x_th := 1 / y_th ** (1 / 4)),
    y_th * (torch.abs(x) < x_th) + torch.nan_to_num(1 / x**4) * (torch.abs(x) >= x_th),
)
f_inv5 = lambda x, y_th: (
    (x_th := 1 / y_th ** (1 / 5)),
    y_th / x_th * x * (torch.abs(x) < x_th)
    + torch.nan_to_num(1 / x**5) * (torch.abs(x) >= x_th),
)
f_sqrt = lambda x, y_th: (
    (x_th := 1 / y_th**2),
    x_th / y_th * x * (torch.abs(x) < x_th)
    + torch.nan_to_num(torch.sqrt(torch.abs(x)) * torch.sign(x))
    * (torch.abs(x) >= x_th),
)
f_power1d5 = lambda x, y_th: torch.abs(x) ** 1.5
f_invsqrt = lambda x, y_th: (
    (x_th := 1 / y_th**2),
    y_th * (torch.abs(x) < x_th)
    + torch.nan_to_num(1 / torch.sqrt(torch.abs(x))) * (torch.abs(x) >= x_th),
)
f_log = lambda x, y_th: (
    (x_th := torch.e ** (-y_th)),
    -y_th * (torch.abs(x) < x_th)
    + torch.nan_to_num(torch.log(torch.abs(x))) * (torch.abs(x) >= x_th),
)
f_tan = lambda x, y_th: (
    (clip := x % torch.pi),
    (delta := torch.pi / 2 - torch.arctan(y_th)),
    -y_th / delta * (clip - torch.pi / 2) * (torch.abs(clip - torch.pi / 2) < delta)
    + torch.nan_to_num(torch.tan(clip)) * (torch.abs(clip - torch.pi / 2) >= delta),
)
f_arctanh = lambda x, y_th: (
    (delta := 1 - torch.tanh(y_th) + 1e-4),
    y_th * torch.sign(x) * (torch.abs(x) > 1 - delta)
    + torch.nan_to_num(torch.arctanh(x)) * (torch.abs(x) <= 1 - delta),
)
f_arcsin = lambda x, y_th: (
    (),
    torch.pi / 2 * torch.sign(x) * (torch.abs(x) > 1)
    + torch.nan_to_num(torch.arcsin(x)) * (torch.abs(x) <= 1),
)
f_arccos = lambda x, y_th: (
    (),
    torch.pi / 2 * (1 - torch.sign(x)) * (torch.abs(x) > 1)
    + torch.nan_to_num(torch.arccos(x)) * (torch.abs(x) <= 1),
)
f_exp = lambda x, y_th: (
    (x_th := torch.log(y_th)),
    y_th * (x > x_th) + torch.exp(x) * (x <= x_th),
)

SYMBOLIC_LIB = {
    "x": (lambda x: x, lambda x: x, 1, lambda x, y_th: ((), x)),
    "x^2": (lambda x: x**2, lambda x: x**2, 2, lambda x, y_th: ((), x**2)),
    "x^3": (lambda x: x**3, lambda x: x**3, 3, lambda x, y_th: ((), x**3)),
    "x^4": (lambda x: x**4, lambda x: x**4, 3, lambda x, y_th: ((), x**4)),
    "x^5": (lambda x: x**5, lambda x: x**5, 3, lambda x, y_th: ((), x**5)),
    "1/x": (lambda x: 1 / x, lambda x: 1 / x, 2, f_inv),
    "1/x^2": (lambda x: 1 / x**2, lambda x: 1 / x**2, 2, f_inv2),
    "1/x^3": (lambda x: 1 / x**3, lambda x: 1 / x**3, 3, f_inv3),
    "1/x^4": (lambda x: 1 / x**4, lambda x: 1 / x**4, 4, f_inv4),
    "1/x^5": (lambda x: 1 / x**5, lambda x: 1 / x**5, 5, f_inv5),
    "sqrt": (lambda x: torch.sqrt(x), lambda x: sympy.sqrt(x), 2, f_sqrt),
    "x^0.5": (lambda x: torch.sqrt(x), lambda x: sympy.sqrt(x), 2, f_sqrt),
    "x^1.5": (
        lambda x: torch.sqrt(x) ** 3,
        lambda x: sympy.sqrt(x) ** 3,
        4,
        f_power1d5,
    ),
    "1/sqrt(x)": (
        lambda x: 1 / torch.sqrt(x),
        lambda x: 1 / sympy.sqrt(x),
        2,
        f_invsqrt,
    ),
    "1/x^0.5": (lambda x: 1 / torch.sqrt(x), lambda x: 1 / sympy.sqrt(x), 2, f_invsqrt),
    "exp": (lambda x: torch.exp(x), lambda x: sympy.exp(x), 2, f_exp),
    "log": (lambda x: torch.log(x), lambda x: sympy.log(x), 2, f_log),
    "abs": (
        lambda x: torch.abs(x),
        lambda x: sympy.Abs(x),
        3,
        lambda x, y_th: ((), torch.abs(x)),
    ),
    "sin": (
        lambda x: torch.sin(x),
        lambda x: sympy.sin(x),
        2,
        lambda x, y_th: ((), torch.sin(x)),
    ),
    "cos": (
        lambda x: torch.cos(x),
        lambda x: sympy.cos(x),
        2,
        lambda x, y_th: ((), torch.cos(x)),
    ),
    "tan": (lambda x: torch.tan(x), lambda x: sympy.tan(x), 3, f_tan),
    "tanh": (
        lambda x: torch.tanh(x),
        lambda x: sympy.tanh(x),
        3,
        lambda x, y_th: ((), torch.tanh(x)),
    ),
    "sgn": (
        lambda x: torch.sign(x),
        lambda x: sympy.sign(x),
        3,
        lambda x, y_th: ((), torch.sign(x)),
    ),
    "arcsin": (lambda x: torch.arcsin(x), lambda x: sympy.asin(x), 4, f_arcsin),
    "arccos": (lambda x: torch.arccos(x), lambda x: sympy.acos(x), 4, f_arccos),
    "arctan": (
        lambda x: torch.arctan(x),
        lambda x: sympy.atan(x),
        4,
        lambda x, y_th: ((), torch.arctan(x)),
    ),
    "arctanh": (lambda x: torch.arctanh(x), lambda x: sympy.atanh(x), 4, f_arctanh),
    "0": (lambda x: x * 0, lambda x: x * 0, 0, lambda x, y_th: ((), x * 0)),
    "gaussian": (
        lambda x: torch.exp(-(x**2)),
        lambda x: sympy.exp(-(x**2)),
        3,
        lambda x, y_th: ((), torch.exp(-(x**2))),
    ),
}


def create_dataset(
    f,
    n_var=2,
    f_mode="col",
    ranges=[-1, 1],
    train_num=1000,
    test_num=1000,
    normalize_input=False,
    normalize_label=False,
    device="cpu",
    seed=0,
):
    """
    Create dataset

    Args:
        f: function
            the symbolic formula used to create the synthetic dataset
        ranges: list or np.array; shape (2,) or (n_var, 2)
            the range of input variables. Default: [-1,1].
        train_num: int
            the number of training samples. Default: 1000.
        test_num: int
            the number of test samples. Default: 1000.
        normalize_input: bool
            If True, apply normalization to inputs. Default: False.
        normalize_label: bool
            If True, apply normalization to labels. Default: False.
        device: str
            device. Default: 'cpu'.
        seed: int
            random seed. Default: 0.

    Returns:
        dataset: dict
            Train/test inputs/labels are dataset['train_input'], dataset['train_label'],
            dataset['test_input'], dataset['test_label']
    """

    np.random.seed(seed)
    torch.manual_seed(seed)

    if len(np.array(ranges).shape) == 1:
        ranges = np.array(ranges * n_var).reshape(n_var, 2)
    else:
        ranges = np.array(ranges)

    train_input = torch.zeros(train_num, n_var)
    test_input = torch.zeros(test_num, n_var)
    for i in range(n_var):
        train_input[:, i] = (
            torch.rand(
                train_num,
            )
            * (ranges[i, 1] - ranges[i, 0])
            + ranges[i, 0]
        )
        test_input[:, i] = (
            torch.rand(
                test_num,
            )
            * (ranges[i, 1] - ranges[i, 0])
            + ranges[i, 0]
        )

    if f_mode == "col":
        train_label = f(train_input)
        test_label = f(test_input)
    elif f_mode == "row":
        train_label = f(train_input.T)
        test_label = f(test_input.T)
    else:
        print(f"f_mode {f_mode} not recognized")

    # if has only 1 dimension
    if len(train_label.shape) == 1:
        train_label = train_label.unsqueeze(dim=1)
        test_label = test_label.unsqueeze(dim=1)

    def normalize(data, mean, std):
        return (data - mean) / std

    if normalize_input:
        mean_input = torch.mean(train_input, dim=0, keepdim=True)
        std_input = torch.std(train_input, dim=0, keepdim=True)
        train_input = normalize(train_input, mean_input, std_input)
        test_input = normalize(test_input, mean_input, std_input)

    if normalize_label:
        mean_label = torch.mean(train_label, dim=0, keepdim=True)
        std_label = torch.std(train_label, dim=0, keepdim=True)
        train_label = normalize(train_label, mean_label, std_label)
        test_label = normalize(test_label, mean_label, std_label)

    dataset = {}
    dataset["train_input"] = train_input.to(device)
    dataset["test_input"] = test_input.to(device)

    dataset["train_label"] = train_label.to(device)
    dataset["test_label"] = test_label.to(device)

    return dataset
